# pylint: disable=too-many-lines
from unittest import TestCase

from lxml import etree

from pcs.common.reports import codes as report_codes
from pcs.lib.cib.resource import common
from pcs.lib.cib.tools import IdProvider

from pcs_test.tools import fixture
from pcs_test.tools.assertions import (
    assert_report_item_list_equal,
    assert_xml_equal,
)
from pcs_test.tools.xml import etree_to_str

fixture_cib = etree.fromstring(
    """
    <cib>
        <configuration>
            <resources>
                <primitive id="A" />
                <clone id="B-clone">
                    <primitive id="B" />
                </clone>
                <master id="C-master">
                    <primitive id="C" />
                </master>
                <group id="D">
                    <primitive id="D1" />
                    <primitive id="D2" />
                </group>
                <clone id="E-clone">
                    <group id="E">
                        <primitive id="E1" />
                        <primitive id="E2" />
                    </group>
                </clone>
                <master id="F-master">
                    <group id="F">
                        <primitive id="F1" />
                        <primitive id="F2" />
                    </group>
                </master>
                <bundle id="G-bundle" />
                <bundle id="H-bundle">
                    <primitive id="H" />
                </bundle>
                <group id="I">
                    <primitive id="I1" />
                </group>
                <clone id="J-clone">
                    <group id="J">
                        <primitive id="J1" />
                    </group>
                </clone>
                <master id="K-master">
                    <group id="K">
                        <primitive id="K1" />
                    </group>
                </master>
            </resources>
        </configuration>
    </cib>
    """
)


class AreMetaDisabled(TestCase):
    def test_detect_is_disabled(self):
        self.assertTrue(common.are_meta_disabled({"target-role": "Stopped"}))
        self.assertTrue(common.are_meta_disabled({"target-role": "stopped"}))

    def test_detect_is_not_disabled(self):
        self.assertFalse(common.are_meta_disabled({}))
        self.assertFalse(common.are_meta_disabled({"target-role": "any"}))


class IsCloneDeactivatedByMeta(TestCase):
    def assert_is_disabled(self, meta_attributes):
        self.assertTrue(common.is_clone_deactivated_by_meta(meta_attributes))

    def assert_is_not_disabled(self, meta_attributes):
        self.assertFalse(common.is_clone_deactivated_by_meta(meta_attributes))

    def test_detect_is_disabled(self):
        self.assert_is_disabled({"target-role": "Stopped"})
        self.assert_is_disabled({"target-role": "stopped"})
        self.assert_is_disabled({"clone-max": "0"})
        self.assert_is_disabled({"clone-max": "00"})
        self.assert_is_disabled({"clone-max": 0})
        self.assert_is_disabled({"clone-node-max": "0"})
        self.assert_is_disabled({"clone-node-max": "abc1"})

    def test_detect_is_not_disabled(self):
        self.assert_is_not_disabled({})
        self.assert_is_not_disabled({"target-role": "any"})
        self.assert_is_not_disabled({"clone-max": "1"})
        self.assert_is_not_disabled({"clone-max": "01"})
        self.assert_is_not_disabled({"clone-max": 1})
        self.assert_is_not_disabled({"clone-node-max": "1"})
        self.assert_is_not_disabled({"clone-node-max": 1})
        self.assert_is_not_disabled({"clone-node-max": "1abc"})
        self.assert_is_not_disabled({"clone-node-max": "1.1"})


class FindOneOrMoreResources(TestCase):
    def setUp(self):
        self.cib = etree.fromstring(
            """
            <resources>
                <primitive id="R1" />
                <primitive id="R2" />
                <primitive id="R3" />
                <primitive id="R1x" />
                <primitive id="R2x" />
            </resources>
        """
        )

        def searcher(resource_element):
            return [
                resource_element.getparent().find(
                    ".//*[@id='{0}x']".format(resource_element.get("id"))
                )
            ]

        self.additional_search = searcher

    def test_one_existing(self):
        resource, report_list = common.find_one_resource(self.cib, "R1")
        self.assertEqual("R1", resource.attrib.get("id"))
        assert_report_item_list_equal(report_list, [])

    def test_one_nonexistent(self):
        resource, report_list = common.find_one_resource(self.cib, "R-missing")
        self.assertIsNone(resource)
        assert_report_item_list_equal(
            report_list,
            [
                fixture.report_not_found("R-missing", context_type="resources"),
            ],
        )

    def test_more_existing(self):
        resource_list, report_list = common.find_resources(
            self.cib, ["R1", "R2"]
        )
        self.assertEqual(
            ["R1", "R2"],
            [resource.attrib.get("id") for resource in resource_list],
        )
        assert_report_item_list_equal(report_list, [])

    def test_more_some_missing(self):
        resource_list, report_list = common.find_resources(
            self.cib, ["R1", "R2", "RY1", "RY2"]
        )
        self.assertEqual(
            ["R1", "R2"],
            [resource.attrib.get("id") for resource in resource_list],
        )
        assert_report_item_list_equal(
            report_list,
            [
                fixture.report_not_found("RY1", context_type="resources"),
                fixture.report_not_found("RY2", context_type="resources"),
            ],
        )


class FindResourcesMixin:
    _iterable_type = list

    def assert_find_resources(self, input_resource_id, output_resource_ids):
        self.assertEqual(
            self._iterable_type(output_resource_ids),
            self._iterable_type(
                [
                    element.get("id", "")
                    for element in self._tested_fn(
                        fixture_cib.find(
                            './/*[@id="{0}"]'.format(input_resource_id)
                        )
                    )
                ]
            ),
        )

    def test_group(self):
        self.assert_find_resources("D", ["D1", "D2"])

    def test_group_in_clone(self):
        self.assert_find_resources("E", ["E1", "E2"])

    def test_group_in_master(self):
        self.assert_find_resources("F", ["F1", "F2"])

    def test_cloned_primitive(self):
        self.assert_find_resources("B-clone", ["B"])

    def test_mastered_primitive(self):
        self.assert_find_resources("C-master", ["C"])

    def test_bundle_empty(self):
        self.assert_find_resources("G-bundle", [])

    def test_bundle_with_primitive(self):
        self.assert_find_resources("H-bundle", ["H"])

    def test_primitive(self):
        raise NotImplementedError()

    def test_primitive_in_clone(self):
        raise NotImplementedError()

    def test_primitive_in_master(self):
        raise NotImplementedError()

    def test_primitive_in_group(self):
        raise NotImplementedError()

    def test_primitive_in_bundle(self):
        raise NotImplementedError()

    def test_cloned_group(self):
        raise NotImplementedError()

    def test_mastered_group(self):
        raise NotImplementedError()


class FindPrimitives(TestCase, FindResourcesMixin):
    _tested_fn = staticmethod(common.find_primitives)

    def test_primitive(self):
        self.assert_find_resources("A", ["A"])

    def test_primitive_in_clone(self):
        self.assert_find_resources("B", ["B"])

    def test_primitive_in_master(self):
        self.assert_find_resources("C", ["C"])

    def test_primitive_in_group(self):
        self.assert_find_resources("D1", ["D1"])
        self.assert_find_resources("D2", ["D2"])
        self.assert_find_resources("E1", ["E1"])
        self.assert_find_resources("E2", ["E2"])
        self.assert_find_resources("F1", ["F1"])
        self.assert_find_resources("F2", ["F2"])

    def test_primitive_in_bundle(self):
        self.assert_find_resources("H", ["H"])

    def test_cloned_group(self):
        self.assert_find_resources("E-clone", ["E1", "E2"])

    def test_mastered_group(self):
        self.assert_find_resources("F-master", ["F1", "F2"])


class GetAllInnerResources(TestCase, FindResourcesMixin):
    _iterable_type = set
    _tested_fn = staticmethod(common.get_all_inner_resources)

    def test_primitive(self):
        self.assert_find_resources("A", set())

    def test_primitive_in_clone(self):
        self.assert_find_resources("B", set())

    def test_primitive_in_master(self):
        self.assert_find_resources("C", set())

    def test_primitive_in_group(self):
        self.assert_find_resources("D1", set())
        self.assert_find_resources("D2", set())
        self.assert_find_resources("E1", set())
        self.assert_find_resources("E2", set())
        self.assert_find_resources("F1", set())
        self.assert_find_resources("F2", set())

    def test_primitive_in_bundle(self):
        self.assert_find_resources("H", set())

    def test_cloned_group(self):
        self.assert_find_resources("E-clone", {"E", "E1", "E2"})

    def test_mastered_group(self):
        self.assert_find_resources("F-master", {"F", "F1", "F2"})


class GetInnerResources(TestCase, FindResourcesMixin):
    _tested_fn = staticmethod(common.get_inner_resources)

    def test_primitive(self):
        self.assert_find_resources("A", [])

    def test_primitive_in_clone(self):
        self.assert_find_resources("B", [])

    def test_primitive_in_master(self):
        self.assert_find_resources("C", [])

    def test_primitive_in_group(self):
        self.assert_find_resources("D1", [])
        self.assert_find_resources("D2", [])
        self.assert_find_resources("E1", [])
        self.assert_find_resources("E2", [])
        self.assert_find_resources("F1", [])
        self.assert_find_resources("F2", [])

    def test_primitive_in_bundle(self):
        self.assert_find_resources("H", [])

    def test_mastered_group(self):
        self.assert_find_resources("F-master", ["F"])

    def test_cloned_group(self):
        self.assert_find_resources("E-clone", ["E"])


class IsWrapperResource(TestCase):
    def assert_is_wrapper(self, res_id, is_wrapper):
        self.assertEqual(
            is_wrapper,
            common.is_wrapper_resource(
                fixture_cib.find('.//*[@id="{0}"]'.format(res_id))
            ),
        )

    def test_primitive(self):
        self.assert_is_wrapper("A", False)

    def test_primitive_in_clone(self):
        self.assert_is_wrapper("B", False)

    def test_primitive_in_master(self):
        self.assert_is_wrapper("C", False)

    def test_primitive_in_group(self):
        self.assert_is_wrapper("D1", False)
        self.assert_is_wrapper("D2", False)
        self.assert_is_wrapper("E1", False)
        self.assert_is_wrapper("E2", False)
        self.assert_is_wrapper("F1", False)
        self.assert_is_wrapper("F2", False)

    def test_primitive_in_bundle(self):
        self.assert_is_wrapper("H", False)

    def test_cloned_group(self):
        self.assert_is_wrapper("E-clone", True)

    def test_mastered_group(self):
        self.assert_is_wrapper("F-master", True)

    def test_group(self):
        self.assert_is_wrapper("D", True)

    def test_group_in_clone(self):
        self.assert_is_wrapper("E", True)

    def test_group_in_master(self):
        self.assert_is_wrapper("F", True)

    def test_cloned_primitive(self):
        self.assert_is_wrapper("B-clone", True)

    def test_mastered_primitive(self):
        self.assert_is_wrapper("C-master", True)

    def test_bundle_empty(self):
        self.assert_is_wrapper("G-bundle", True)

    def test_bundle_with_primitive(self):
        self.assert_is_wrapper("H-bundle", True)


class GetParentResource(TestCase):
    def assert_parent_resource(self, input_resource_id, output_resource_id):
        res_el = common.get_parent_resource(
            fixture_cib.find('.//*[@id="{0}"]'.format(input_resource_id))
        )
        self.assertEqual(
            output_resource_id, res_el.get("id") if res_el is not None else None
        )

    def test_primitive(self):
        self.assert_parent_resource("A", None)

    def test_primitive_in_clone(self):
        self.assert_parent_resource("B", "B-clone")

    def test_primitive_in_master(self):
        self.assert_parent_resource("C", "C-master")

    def test_primitive_in_group(self):
        self.assert_parent_resource("D1", "D")
        self.assert_parent_resource("D2", "D")
        self.assert_parent_resource("E1", "E")
        self.assert_parent_resource("E2", "E")
        self.assert_parent_resource("F1", "F")
        self.assert_parent_resource("F2", "F")

    def test_primitive_in_bundle(self):
        self.assert_parent_resource("H", "H-bundle")

    def test_cloned_group(self):
        self.assert_parent_resource("E-clone", None)

    def test_mastered_group(self):
        self.assert_parent_resource("F-master", None)

    def test_group(self):
        self.assert_parent_resource("D", None)

    def test_group_in_clone(self):
        self.assert_parent_resource("E", "E-clone")

    def test_group_in_master(self):
        self.assert_parent_resource("F", "F-master")

    def test_cloned_primitive(self):
        self.assert_parent_resource("B-clone", None)

    def test_mastered_primitive(self):
        self.assert_parent_resource("C-master", None)

    def test_bundle_empty(self):
        self.assert_parent_resource("G-bundle", None)

    def test_bundle_with_primitive(self):
        self.assert_parent_resource("H-bundle", None)


class FindResourcesToEnable(TestCase):
    def assert_find_resources(self, input_resource_id, output_resource_ids):
        self.assertEqual(
            output_resource_ids,
            [
                element.get("id", "")
                for element in common.find_resources_to_enable(
                    fixture_cib.find(
                        './/*[@id="{0}"]'.format(input_resource_id)
                    )
                )
            ],
        )

    def test_primitive(self):
        self.assert_find_resources("A", ["A"])

    def test_primitive_in_clone(self):
        self.assert_find_resources("B", ["B", "B-clone"])

    def test_primitive_in_master(self):
        self.assert_find_resources("C", ["C", "C-master"])

    def test_primitive_in_group(self):
        self.assert_find_resources("D1", ["D1"])
        self.assert_find_resources("D2", ["D2"])
        self.assert_find_resources("E1", ["E1"])
        self.assert_find_resources("E2", ["E2"])
        self.assert_find_resources("F1", ["F1"])
        self.assert_find_resources("F2", ["F2"])

    def test_primitive_in_bundle(self):
        self.assert_find_resources("H", ["H", "H-bundle"])

    def test_group(self):
        self.assert_find_resources("D", ["D"])

    def test_group_in_clone(self):
        self.assert_find_resources("E", ["E", "E-clone"])

    def test_group_in_master(self):
        self.assert_find_resources("F", ["F", "F-master"])

    def test_cloned_primitive(self):
        self.assert_find_resources("B-clone", ["B-clone", "B"])

    def test_cloned_group(self):
        self.assert_find_resources("E-clone", ["E-clone", "E"])

    def test_mastered_primitive(self):
        self.assert_find_resources("C-master", ["C-master", "C"])

    def test_mastered_group(self):
        self.assert_find_resources("F-master", ["F-master", "F"])

    def test_bundle_empty(self):
        self.assert_find_resources("G-bundle", ["G-bundle"])

    def test_bundle_with_primitive(self):
        self.assert_find_resources("H-bundle", ["H-bundle", "H"])


class Enable(TestCase):
    @staticmethod
    def assert_enabled(pre, post):
        resource = etree.fromstring(pre)
        common.enable(resource, IdProvider(resource))
        assert_xml_equal(post, etree_to_str(resource))

    def test_disabled(self):
        self.assert_enabled(
            """
                <resource>
                    <meta_attributes>
                        <nvpair name="target-role" value="something" />
                    </meta_attributes>
                </resource>
            """,
            """
                <resource>
                    <meta_attributes />
                </resource>
            """,
        )

    def test_enabled(self):
        self.assert_enabled(
            """
                <resource>
                </resource>
            """,
            """
                <resource>
                </resource>
            """,
        )

    def test_only_first_meta(self):
        # this captures the current behavior
        # once pcs supports more instance and meta attributes for each resource,
        # this test should be reconsidered
        self.assert_enabled(
            """
                <resource>
                    <meta_attributes id="meta1">
                        <nvpair name="target-role" value="something" />
                    </meta_attributes>
                    <meta_attributes id="meta2">
                        <nvpair name="target-role" value="something" />
                    </meta_attributes>
                </resource>
            """,
            """
                <resource>
                    <meta_attributes id="meta1" />
                    <meta_attributes id="meta2">
                        <nvpair name="target-role" value="something" />
                    </meta_attributes>
                </resource>
            """,
        )


class Disable(TestCase):
    @staticmethod
    def assert_disabled(pre, post):
        resource = etree.fromstring(pre)
        common.disable(resource, IdProvider(resource))
        assert_xml_equal(post, etree_to_str(resource))

    def test_disabled(self):
        xml = """
            <resource id="R">
                <meta_attributes id="R-meta_attributes">
                    <nvpair id="R-meta_attributes-target-role"
                        name="target-role" value="Stopped" />
                </meta_attributes>
            </resource>
        """
        self.assert_disabled(xml, xml)

    def test_enabled(self):
        self.assert_disabled(
            """
                <resource id="R">
                </resource>
            """,
            """
                <resource id="R">
                    <meta_attributes id="R-meta_attributes">
                        <nvpair id="R-meta_attributes-target-role"
                            name="target-role" value="Stopped" />
                    </meta_attributes>
                </resource>
            """,
        )

    def test_only_first_meta(self):
        # this captures the current behavior
        # once pcs supports more instance and meta attributes for each resource,
        # this test should be reconsidered
        self.assert_disabled(
            """
                <resource id="R">
                    <meta_attributes id="R-meta_attributes">
                    </meta_attributes>
                    <meta_attributes id="R-meta_attributes-2">
                    </meta_attributes>
                </resource>
            """,
            """
                <resource id="R">
                    <meta_attributes id="R-meta_attributes">
                        <nvpair id="R-meta_attributes-target-role"
                            name="target-role" value="Stopped" />
                    </meta_attributes>
                    <meta_attributes id="R-meta_attributes-2">
                    </meta_attributes>
                </resource>
            """,
        )


class FindResourcesToManage(TestCase):
    def assert_find_resources(self, input_resource_id, output_resource_ids):
        self.assertEqual(
            output_resource_ids,
            [
                element.get("id", "")
                for element in common.find_resources_to_manage(
                    fixture_cib.find(
                        './/*[@id="{0}"]'.format(input_resource_id)
                    )
                )
            ],
        )

    def test_primitive(self):
        self.assert_find_resources("A", ["A"])

    def test_primitive_in_clone(self):
        self.assert_find_resources("B", ["B", "B-clone"])

    def test_primitive_in_master(self):
        self.assert_find_resources("C", ["C", "C-master"])

    def test_primitive_in_group(self):
        self.assert_find_resources("D1", ["D1", "D"])
        self.assert_find_resources("D2", ["D2", "D"])
        self.assert_find_resources("E1", ["E1", "E-clone", "E"])
        self.assert_find_resources("E2", ["E2", "E-clone", "E"])
        self.assert_find_resources("F1", ["F1", "F-master", "F"])
        self.assert_find_resources("F2", ["F2", "F-master", "F"])

    def test_primitive_in_bundle(self):
        self.assert_find_resources("H", ["H", "H-bundle"])

    def test_group(self):
        self.assert_find_resources("D", ["D", "D1", "D2"])

    def test_group_in_clone(self):
        self.assert_find_resources("E", ["E", "E-clone", "E1", "E2"])

    def test_group_in_master(self):
        self.assert_find_resources("F", ["F", "F-master", "F1", "F2"])

    def test_cloned_primitive(self):
        self.assert_find_resources("B-clone", ["B-clone", "B"])

    def test_cloned_group(self):
        self.assert_find_resources("E-clone", ["E-clone", "E", "E1", "E2"])

    def test_mastered_primitive(self):
        self.assert_find_resources("C-master", ["C-master", "C"])

    def test_mastered_group(self):
        self.assert_find_resources("F-master", ["F-master", "F", "F1", "F2"])

    def test_bundle_empty(self):
        self.assert_find_resources("G-bundle", ["G-bundle"])

    def test_bundle_with_primitive(self):
        self.assert_find_resources("H-bundle", ["H-bundle", "H"])


class FindResourcesToUnmanage(TestCase):
    def assert_find_resources(self, input_resource_id, output_resource_ids):
        self.assertEqual(
            output_resource_ids,
            [
                element.get("id", "")
                for element in common.find_resources_to_unmanage(
                    fixture_cib.find(
                        './/*[@id="{0}"]'.format(input_resource_id)
                    )
                )
            ],
        )

    def test_primitive(self):
        self.assert_find_resources("A", ["A"])

    def test_primitive_in_clone(self):
        self.assert_find_resources("B", ["B"])

    def test_primitive_in_master(self):
        self.assert_find_resources("C", ["C"])

    def test_primitive_in_group(self):
        self.assert_find_resources("D1", ["D1"])
        self.assert_find_resources("D2", ["D2"])
        self.assert_find_resources("E1", ["E1"])
        self.assert_find_resources("E2", ["E2"])
        self.assert_find_resources("F1", ["F1"])
        self.assert_find_resources("F2", ["F2"])

    def test_primitive_in_bundle(self):
        self.assert_find_resources("H", ["H"])

    def test_group(self):
        self.assert_find_resources("D", ["D1", "D2"])

    def test_group_in_clone(self):
        self.assert_find_resources("E", ["E1", "E2"])

    def test_group_in_master(self):
        self.assert_find_resources("F", ["F1", "F2"])

    def test_cloned_primitive(self):
        self.assert_find_resources("B-clone", ["B"])

    def test_cloned_group(self):
        self.assert_find_resources("E-clone", ["E1", "E2"])

    def test_mastered_primitive(self):
        self.assert_find_resources("C-master", ["C"])

    def test_mastered_group(self):
        self.assert_find_resources("F-master", ["F1", "F2"])

    def test_bundle_empty(self):
        self.assert_find_resources("G-bundle", ["G-bundle"])

    def test_bundle_with_primitive(self):
        self.assert_find_resources("H-bundle", ["H-bundle", "H"])


class Manage(TestCase):
    @staticmethod
    def assert_managed(pre, post):
        resource = etree.fromstring(pre)
        common.manage(resource, IdProvider(resource))
        assert_xml_equal(post, etree_to_str(resource))

    def test_unmanaged(self):
        self.assert_managed(
            """
                <resource>
                    <meta_attributes>
                        <nvpair name="is-managed" value="something" />
                    </meta_attributes>
                </resource>
            """,
            """
                <resource>
                    <meta_attributes />
                </resource>
            """,
        )

    def test_managed(self):
        self.assert_managed(
            """
                <resource>
                </resource>
            """,
            """
                <resource>
                </resource>
            """,
        )

    def test_only_first_meta(self):
        # this captures the current behavior
        # once pcs supports more instance and meta attributes for each resource,
        # this test should be reconsidered
        self.assert_managed(
            """
                <resource>
                    <meta_attributes id="meta1">
                        <nvpair name="is-managed" value="something" />
                    </meta_attributes>
                    <meta_attributes id="meta2">
                        <nvpair name="is-managed" value="something" />
                    </meta_attributes>
                </resource>
            """,
            """
                <resource>
                    <meta_attributes id="meta1" />
                    <meta_attributes id="meta2">
                        <nvpair name="is-managed" value="something" />
                    </meta_attributes>
                </resource>
            """,
        )


class Unmanage(TestCase):
    @staticmethod
    def assert_unmanaged(pre, post):
        resource = etree.fromstring(pre)
        common.unmanage(resource, IdProvider(resource))
        assert_xml_equal(post, etree_to_str(resource))

    def test_unmanaged(self):
        xml = """
            <resource id="R">
                <meta_attributes id="R-meta_attributes">
                    <nvpair id="R-meta_attributes-is-managed"
                        name="is-managed" value="false" />
                </meta_attributes>
            </resource>
        """
        self.assert_unmanaged(xml, xml)

    def test_managed(self):
        self.assert_unmanaged(
            """
                <resource id="R">
                </resource>
            """,
            """
                <resource id="R">
                    <meta_attributes id="R-meta_attributes">
                        <nvpair id="R-meta_attributes-is-managed"
                            name="is-managed" value="false" />
                    </meta_attributes>
                </resource>
            """,
        )

    def test_only_first_meta(self):
        # this captures the current behavior
        # once pcs supports more instance and meta attributes for each resource,
        # this test should be reconsidered
        self.assert_unmanaged(
            """
                <resource id="R">
                    <meta_attributes id="R-meta_attributes">
                    </meta_attributes>
                    <meta_attributes id="R-meta_attributes-2">
                    </meta_attributes>
                </resource>
            """,
            """
                <resource id="R">
                    <meta_attributes id="R-meta_attributes">
                        <nvpair id="R-meta_attributes-is-managed"
                            name="is-managed" value="false" />
                    </meta_attributes>
                    <meta_attributes id="R-meta_attributes-2">
                    </meta_attributes>
                </resource>
            """,
        )


class ValidateMoveBanClearMixin:
    # pylint: disable=too-many-public-methods,line-too-long
    @staticmethod
    def _fixture_clone(promotable=False):
        return etree.fromstring(
            f"""
            <clone id="R-clone">
                <primitive id="R" />
                <meta_attributes>
                    <nvpair name="promotable" value="{'true' if promotable else 'false'}" />
                </meta_attributes>
            </clone>
            """
        )

    @staticmethod
    def _fixture_group_clone(promotable=False):
        return etree.fromstring(
            f"""
            <clone id="G-clone">
                <group id="G">
                    <primitive id="R" />
                </group>
                <meta_attributes>
                    <nvpair name="promotable" value="{'true' if promotable else 'false'}" />
                </meta_attributes>
            </clone>
            """
        )

    @staticmethod
    def _fixture_master():
        return etree.fromstring(
            """
            <master id="R-master">
                <primitive id="R" />
            </master>
            """
        )

    @staticmethod
    def _fixture_group_master():
        return etree.fromstring(
            """
            <master id="G-master">
                <group id="G">
                    <primitive id="R" />
                </group>
            </master>
            """
        )

    def test_master_true_promotable_clone(self):
        element = self._fixture_clone(True)
        assert_report_item_list_equal(self.validate(element, True), [])

    def test_master_false_promotable_clone(self):
        element = self._fixture_clone(True)
        assert_report_item_list_equal(self.validate(element, False), [])

    def test_master_true_clone(self):
        element = self._fixture_clone(False)
        assert_report_item_list_equal(
            self.validate(element, True),
            [
                fixture.error(
                    self.report_code_bad_master,
                    resource_id="R-clone",
                    promotable_id=None,
                ),
            ],
        )

    def test_master_false_clone(self):
        element = self._fixture_clone(False)
        assert_report_item_list_equal(self.validate(element, False), [])

    def test_master_true_master(self):
        element = self._fixture_master()
        assert_report_item_list_equal(self.validate(element, True), [])

    def test_master_false_master(self):
        element = self._fixture_master()
        assert_report_item_list_equal(self.validate(element, False), [])

    def test_master_true_promotable_clone_resource(self):
        element = self._fixture_clone(True)
        assert_report_item_list_equal(
            self.validate(element.find("./primitive"), True),
            [
                fixture.error(
                    self.report_code_bad_master,
                    resource_id="R",
                    promotable_id="R-clone",
                ),
            ],
        )

    def test_master_false_promotable_clone_resource(self):
        element = self._fixture_clone(True)
        assert_report_item_list_equal(
            self.validate(element.find("./primitive"), False), []
        )

    def test_master_true_promotable_clone_group(self):
        element = self._fixture_group_clone(True)
        assert_report_item_list_equal(
            self.validate(element.find("./group"), True),
            [
                fixture.error(
                    self.report_code_bad_master,
                    resource_id="G",
                    promotable_id="G-clone",
                ),
            ],
        )

    def test_master_false_promotable_clone_group(self):
        element = self._fixture_group_clone(True)
        assert_report_item_list_equal(
            self.validate(element.find("./group"), False), []
        )

    def test_master_true_promotable_clone_group_resource(self):
        element = self._fixture_group_clone(True)
        assert_report_item_list_equal(
            self.validate(element.find("./group/primitive"), True),
            [
                fixture.error(
                    self.report_code_bad_master,
                    resource_id="R",
                    promotable_id="G-clone",
                ),
            ],
        )

    def test_master_false_promotable_clone_group_resource(self):
        element = self._fixture_group_clone(True)
        assert_report_item_list_equal(
            self.validate(element.find("./group/primitive"), False), []
        )

    def test_master_true_clone_resource(self):
        element = self._fixture_clone(False)
        assert_report_item_list_equal(
            self.validate(element.find("./primitive"), True),
            [
                fixture.error(
                    self.report_code_bad_master,
                    resource_id="R",
                    promotable_id=None,
                ),
            ],
        )

    def test_master_false_clone_resource(self):
        element = self._fixture_clone(False)
        assert_report_item_list_equal(
            self.validate(element.find("./primitive"), False), []
        )

    def test_master_true_clone_group(self):
        element = self._fixture_group_clone(False)
        assert_report_item_list_equal(
            self.validate(element.find("./group"), True),
            [
                fixture.error(
                    self.report_code_bad_master,
                    resource_id="G",
                    promotable_id=None,
                ),
            ],
        )

    def test_master_false_clone_group(self):
        element = self._fixture_group_clone(False)
        assert_report_item_list_equal(
            self.validate(element.find("./group"), False), []
        )

    def test_master_true_clone_group_resource(self):
        element = self._fixture_group_clone(False)
        assert_report_item_list_equal(
            self.validate(element.find("./group/primitive"), True),
            [
                fixture.error(
                    self.report_code_bad_master,
                    resource_id="R",
                    promotable_id=None,
                ),
            ],
        )

    def test_master_false_clone_group_resource(self):
        element = self._fixture_group_clone(False)
        assert_report_item_list_equal(
            self.validate(element.find("./group/primitive"), False), []
        )

    def test_master_true_master_resource(self):
        element = self._fixture_master()
        assert_report_item_list_equal(
            self.validate(element.find("./primitive"), True),
            [
                fixture.error(
                    self.report_code_bad_master,
                    resource_id="R",
                    promotable_id="R-master",
                ),
            ],
        )

    def test_master_true_master_group(self):
        element = self._fixture_group_master()
        assert_report_item_list_equal(
            self.validate(element.find("./group"), True),
            [
                fixture.error(
                    self.report_code_bad_master,
                    resource_id="G",
                    promotable_id="G-master",
                ),
            ],
        )

    def test_master_true_master_group_resource(self):
        element = self._fixture_group_master()
        assert_report_item_list_equal(
            self.validate(element.find("./group/primitive"), True),
            [
                fixture.error(
                    self.report_code_bad_master,
                    resource_id="R",
                    promotable_id="G-master",
                ),
            ],
        )


class ValidateMove(ValidateMoveBanClearMixin, TestCase):
    validate = staticmethod(common.validate_move)
    report_code_bad_master = (
        report_codes.CANNOT_MOVE_RESOURCE_MASTER_RESOURCE_NOT_PROMOTABLE
    )

    @staticmethod
    def _fixture_bundle():
        return etree.fromstring(
            """
            <bundle id="R-bundle">
                <primitive id="R" />
            </bundle>
            """
        )

    def test_master_false_promotable_clone(self):
        element = self._fixture_clone(True)
        assert_report_item_list_equal(
            self.validate(element, False),
            [],
        )

    def test_master_true_promotable_clone(self):
        element = self._fixture_clone(True)
        assert_report_item_list_equal(
            self.validate(element, True),
            [],
        )

    def test_master_true_clone(self):
        element = self._fixture_clone(False)
        assert_report_item_list_equal(
            self.validate(element, True),
            [
                fixture.error(
                    report_codes.CANNOT_MOVE_RESOURCE_CLONE,
                    resource_id="R-clone",
                ),
            ],
        )

    def test_master_false_clone(self):
        element = self._fixture_clone(False)
        assert_report_item_list_equal(
            self.validate(element, False),
            [
                fixture.error(
                    report_codes.CANNOT_MOVE_RESOURCE_CLONE,
                    resource_id="R-clone",
                ),
            ],
        )

    def test_master_false_master(self):
        element = self._fixture_master()
        assert_report_item_list_equal(
            self.validate(element, False),
            [],
        )

    def test_master_true_master(self):
        element = self._fixture_master()
        assert_report_item_list_equal(
            self.validate(element, True),
            [],
        )

    def test_master_false_promotable_clone_resource(self):
        element = self._fixture_clone(True)
        assert_report_item_list_equal(
            self.validate(element.find("./primitive"), False),
            [
                fixture.error(
                    report_codes.CANNOT_MOVE_RESOURCE_PROMOTABLE_INNER,
                    resource_id="R",
                    promotable_id="R-clone",
                ),
            ],
        )

    def test_master_false_promotable_clone_group(self):
        element = self._fixture_group_clone(True)
        assert_report_item_list_equal(
            self.validate(element.find("./group"), False),
            [
                fixture.error(
                    report_codes.CANNOT_MOVE_RESOURCE_PROMOTABLE_INNER,
                    resource_id="G",
                    promotable_id="G-clone",
                ),
            ],
        )

    def test_master_false_promotable_clone_group_resource(self):
        element = self._fixture_group_clone(True)
        assert_report_item_list_equal(
            self.validate(element.find("./group/primitive"), False),
            [
                fixture.error(
                    report_codes.CANNOT_MOVE_RESOURCE_PROMOTABLE_INNER,
                    resource_id="R",
                    promotable_id="G-clone",
                ),
            ],
        )

    def test_master_true_clone_resource(self):
        element = self._fixture_clone(False)
        assert_report_item_list_equal(
            self.validate(element.find("./primitive"), True),
            [
                fixture.error(
                    report_codes.CANNOT_MOVE_RESOURCE_CLONE,
                    resource_id="R",
                ),
            ],
        )

    def test_master_false_clone_resource(self):
        element = self._fixture_clone(False)
        assert_report_item_list_equal(
            self.validate(element.find("./primitive"), False),
            [
                fixture.error(
                    report_codes.CANNOT_MOVE_RESOURCE_CLONE,
                    resource_id="R",
                ),
            ],
        )

    def test_master_true_clone_group(self):
        element = self._fixture_group_clone(False)
        assert_report_item_list_equal(
            self.validate(element.find("./group"), True),
            [
                fixture.error(
                    report_codes.CANNOT_MOVE_RESOURCE_CLONE,
                    resource_id="G",
                ),
            ],
        )

    def test_master_false_clone_group(self):
        element = self._fixture_group_clone(False)
        assert_report_item_list_equal(
            self.validate(element.find("./group"), False),
            [
                fixture.error(
                    report_codes.CANNOT_MOVE_RESOURCE_CLONE,
                    resource_id="G",
                ),
            ],
        )

    def test_master_true_clone_group_resource(self):
        element = self._fixture_group_clone(False)
        assert_report_item_list_equal(
            self.validate(element.find("./group/primitive"), True),
            [
                fixture.error(
                    report_codes.CANNOT_MOVE_RESOURCE_CLONE,
                    resource_id="R",
                ),
            ],
        )

    def test_master_false_clone_group_resource(self):
        element = self._fixture_group_clone(False)
        assert_report_item_list_equal(
            self.validate(element.find("./group/primitive"), False),
            [
                fixture.error(
                    report_codes.CANNOT_MOVE_RESOURCE_CLONE,
                    resource_id="R",
                ),
            ],
        )

    def test_bundle(self):
        element = self._fixture_bundle()
        assert_report_item_list_equal(
            self.validate(element, False),
            [
                fixture.error(
                    report_codes.CANNOT_MOVE_RESOURCE_BUNDLE,
                    resource_id="R-bundle",
                ),
            ],
        )

    def test_bundle_resource(self):
        element = self._fixture_bundle()
        assert_report_item_list_equal(
            self.validate(element.find("./primitive"), False), []
        )


class ValidateBan(ValidateMoveBanClearMixin, TestCase):
    validate = staticmethod(common.validate_ban)
    report_code_bad_master = (
        report_codes.CANNOT_BAN_RESOURCE_MASTER_RESOURCE_NOT_PROMOTABLE
    )


class ValidateUnmoveUnban(ValidateMoveBanClearMixin, TestCase):
    validate = staticmethod(common.validate_unmove_unban)
    report_code_bad_master = (
        report_codes.CANNOT_UNMOVE_UNBAN_RESOURCE_MASTER_RESOURCE_NOT_PROMOTABLE
    )


class FindResourcesToDelete(TestCase):
    # pylint: disable=too-many-public-methods
    def assert_element2element_list(self, element_id, element_id_list):
        self.assertEqual(
            common.find_resources_to_delete(
                fixture_cib.xpath(f'.//*[@id="{element_id}"]')[0]
            ),
            [
                fixture_cib.xpath(f'.//*[@id="{_id}"]')[0]
                for _id in element_id_list
            ],
        )

    def test_primitive(self):
        self.assert_element2element_list("A", ["A"])

    def test_clone(self):
        self.assert_element2element_list("B-clone", ["B-clone", "B"])

    def test_primitive_in_clone(self):
        self.assert_element2element_list("B", ["B-clone", "B"])

    def test_master(self):
        self.assert_element2element_list("C-master", ["C-master", "C"])

    def test_primitive_in_master(self):
        self.assert_element2element_list("C", ["C-master", "C"])

    def test_group(self):
        self.assert_element2element_list("D", ["D", "D1", "D2"])

    def test_primitive_in_group(self):
        self.assert_element2element_list("D1", ["D1"])

    def test_clone_with_group(self):
        self.assert_element2element_list(
            "E-clone",
            ["E-clone", "E", "E1", "E2"],
        )

    def test_group_in_clone(self):
        self.assert_element2element_list("E", ["E-clone", "E", "E1", "E2"])

    def test_primitive_in_cloned_group(self):
        self.assert_element2element_list("E2", ["E2"])

    def test_master_with_group(self):
        self.assert_element2element_list(
            "F-master",
            ["F-master", "F", "F1", "F2"],
        )

    def test_group_in_master(self):
        self.assert_element2element_list("F", ["F-master", "F", "F1", "F2"])

    def test_primitive_in_mastered_group(self):
        self.assert_element2element_list("F1", ["F1"])

    def test_empty_bundle(self):
        self.assert_element2element_list("G-bundle", ["G-bundle"])

    def test_bundle_with_primitive(self):
        self.assert_element2element_list("H-bundle", ["H-bundle", "H"])

    def test_primitive_in_bundle(self):
        self.assert_element2element_list("H", ["H"])

    def test_group_with_single_primitive(self):
        self.assert_element2element_list("I", ["I", "I1"])

    def test_single_primitive_in_group(self):
        self.assert_element2element_list("I1", ["I", "I1"])

    def test_clone_with_group_with_single_primitive(self):
        self.assert_element2element_list("J-clone", ["J-clone", "J", "J1"])

    def test_group_with_single_primitive_in_clone(self):
        self.assert_element2element_list("J", ["J-clone", "J", "J1"])

    def test_single_primitive_in_cloned_group(self):
        self.assert_element2element_list("J1", ["J-clone", "J", "J1"])

    def test_master_with_group_with_single_primitive(self):
        self.assert_element2element_list("K-master", ["K-master", "K", "K1"])

    def test_group_with_single_primitive_in_master(self):
        self.assert_element2element_list("K", ["K-master", "K", "K1"])

    def test_single_primitive_in_mastered_group(self):
        self.assert_element2element_list("K1", ["K-master", "K", "K1"])



from unittest import TestCase

from pcs.lib.resource_agent import (
    InvalidResourceAgentName,
    ResourceAgentMetadata,
    ResourceAgentName,
    const,
    split_resource_agent_name,
)
from pcs.lib.resource_agent.name import name_to_void_metadata


class SplitResourceAgentName(TestCase):
    def test_returns_resource_agent_name_when_is_valid(self):
        self.assertEqual(
            ResourceAgentName("ocf", "heartbeat", "Dummy"),
            split_resource_agent_name("ocf:heartbeat:Dummy"),
        )

    def test_refuses_string_if_is_not_valid(self):
        self.assertRaises(
            InvalidResourceAgentName,
            lambda: split_resource_agent_name("invalid:resource:agent:string"),
        )

    def test_refuses_with_unknown_standard(self):
        self.assertRaises(
            InvalidResourceAgentName,
            lambda: split_resource_agent_name("unknown:Dummy"),
        )

    def test_refuses_ocf_agent_name_without_provider(self):
        self.assertRaises(
            InvalidResourceAgentName,
            lambda: split_resource_agent_name("ocf:Dummy"),
        )

    def test_refuses_non_ocf_agent_name_with_provider(self):
        self.assertRaises(
            InvalidResourceAgentName,
            lambda: split_resource_agent_name("lsb:provider:Dummy"),
        )

    def test_returns_resource_agent_containing_systemd(self):
        self.assertEqual(
            ResourceAgentName("systemd", None, "lvm2-pvscan"),
            split_resource_agent_name("systemd:lvm2-pvscan"),
        )

    def test_returns_resource_agent_containing_systemd_instance(self):
        self.assertEqual(
            ResourceAgentName("systemd", None, "lvm2-pvscan@252:2"),
            split_resource_agent_name("systemd:lvm2-pvscan@252:2"),
        )

    def test_returns_resource_agent_containing_service(self):
        self.assertEqual(
            ResourceAgentName("service", None, "lvm2-pvscan"),
            split_resource_agent_name("service:lvm2-pvscan"),
        )

    def test_returns_resource_agent_containing_service_instance(self):
        self.assertEqual(
            ResourceAgentName("service", None, "lvm2-pvscan@252:2"),
            split_resource_agent_name("service:lvm2-pvscan@252:2"),
        )

    def test_returns_resource_agent_containing_systemd_instance_short(self):
        self.assertEqual(
            ResourceAgentName("service", None, "getty@tty1"),
            split_resource_agent_name("service:getty@tty1"),
        )

    def test_refuses_systemd_agent_name_with_provider(self):
        self.assertRaises(
            InvalidResourceAgentName,
            lambda: split_resource_agent_name(
                "systemd:provider:lvm2-pvscan252"
            ),
        )

    def test_refuses_systemd_agent_name_with_provider_and_instance(self):
        self.assertRaises(
            InvalidResourceAgentName,
            lambda: split_resource_agent_name(
                "systemd:provider:lvm2-pvscan252@252:2"
            ),
        )

    def test_refuses_service_agent_name_with_provider(self):
        self.assertRaises(
            InvalidResourceAgentName,
            lambda: split_resource_agent_name(
                "service:provider:lvm2-pvscan252"
            ),
        )

    def test_refuses_service_agent_name_with_provider_and_instance(self):
        self.assertRaises(
            InvalidResourceAgentName,
            lambda: split_resource_agent_name(
                "service:provider:lvm2-pvscan252@252:2"
            ),
        )


class NameToDataclass(TestCase):
    def test_success(self):
        name = ResourceAgentName("ocf", "pacemaker", "Dummy")
        self.assertEqual(
            name_to_void_metadata(name),
            ResourceAgentMetadata(
                name,
                False,
                const.OCF_1_0,
                shortdesc=None,
                longdesc=None,
                parameters=[],
                actions=[],
            ),
        )



import json

from pcs import settings

from pcs_test.tools.command_env.mock_node_communicator import (
    place_communication,
    place_multinode_call,
)
from pcs_test.tools.misc import outdent


class HostShortcuts:
    def __init__(self, calls):
        self.__calls = calls

    def check_reachability(
        self,
        node_labels=None,
        communication_list=None,
        name="http.host.check_reachability",
    ):
        """
        Create a call for checking if pcsd is reachable on hosts

        node_labels list -- create success responses from these nodes
        communication_list list -- create custom responses
        name string -- the key of this call
        """
        place_multinode_call(
            self.__calls,
            name,
            node_labels,
            communication_list,
            action="remote/check_auth",
            output='{"success":true}',
            param_list=[("check_auth_only", 1)],
        )

    def check_auth(
        self,
        node_labels=None,
        communication_list=None,
        name="http.host.check_auth",
    ):
        """
        Create a call for checking authentication on hosts

        node_labels list -- create success responses from these nodes
        communication_list list -- create custom responses
        name string -- the key of this call
        """
        place_multinode_call(
            self.__calls,
            name,
            node_labels,
            communication_list,
            action="remote/check_auth",
            output='{"success":true}',
            param_list=[("check_auth_only", 1)],
        )

    def get_host_info(
        self,
        node_labels=None,
        output_data=None,
        communication_list=None,
        name="http.host.get_host_info",
    ):
        """
        Create a call for getting overall info about a host

        node_labels list -- create success responses from these nodes
        output_data dict -- default output data which will be converted to JSON
        communication_list list -- create custom responses
        name string -- the key of this call
        """
        place_multinode_call(
            self.__calls,
            name,
            node_labels,
            communication_list,
            action="remote/check_host",
            output=json.dumps(output_data) if output_data else "",
        )

    def cluster_destroy(
        self,
        node_labels=None,
        communication_list=None,
        name="http.host.cluster_destroy",
    ):
        """
        Create a call for destroying a cluster on the hosts

        node_labels list -- create success responses from these nodes
        communication_list list -- create custom responses
        name string -- the key of this call
        """
        place_multinode_call(
            self.__calls,
            name,
            node_labels,
            communication_list,
            action="remote/cluster_destroy",
        )

    def update_known_hosts(
        self,
        node_labels=None,
        to_add=None,
        to_add_hosts=None,
        communication_list=None,
        name="http.host.update_known_hosts",
    ):
        """
        Create a call for updating known hosts on the hosts.

        node_labels list -- create success responses from these nodes
        dict to_add -- records to add:
                {host_name: {dest_list: [{"addr": , "port": ,}]}}
        list to_add_hosts -- constructs to_add from host names
        communication_list list -- create custom responses
        name string -- the key of this call
        """
        if to_add_hosts and to_add:
            raise AssertionError(
                "Cannot specify both 'to_add_hosts' and 'to_add'"
            )
        if to_add_hosts:
            to_add = {
                name: {
                    "dest_list": [
                        {"addr": name, "port": settings.pcsd_default_port}
                    ]
                }
                for name in to_add_hosts
            }
        add_with_token = {
            name: dict(data, token=None) for name, data in to_add.items()
        }
        place_multinode_call(
            self.__calls,
            name,
            node_labels,
            communication_list,
            action="remote/known_hosts_change",
            param_list=[
                (
                    "data_json",
                    json.dumps(
                        dict(
                            known_hosts_add=add_with_token,
                            known_hosts_remove={},
                        )
                    ),
                )
            ],
        )

    def send_pcsd_cert(
        self,
        cert,
        key,
        node_labels=None,
        communication_list=None,
        name="http.host.send_pcsd_cert",
        before=None,
    ):
        """
        Create a call for sending pcsd SSL cert and key

        string cert -- pcsd SSL certificate
        string key -- pcsd SSL key
        node_labels list -- create success responses from these nodes
        communication_list list -- create custom responses
        name string -- the key of this call
        """
        place_multinode_call(
            self.__calls,
            name,
            node_labels,
            communication_list,
            action="remote/set_certs",
            param_list=[("ssl_cert", cert), ("ssl_key", key)],
            before=before,
        )

    def enable_cluster(
        self,
        node_labels=None,
        communication_list=None,
        name="http.host.enable_cluster",
    ):
        """
        Create a call for enabling cluster on the nodes.

        node_labels list -- create success responses from these nodes
        communication_list list -- create custom responses
        name string -- the key of this call
        """
        place_multinode_call(
            self.__calls,
            name,
            node_labels,
            communication_list,
            action="remote/cluster_enable",
        )

    def start_cluster(
        self,
        node_labels=None,
        communication_list=None,
        name="http.host.start_cluster",
    ):
        """
        Create a call for starting cluster on the nodes.

        node_labels list -- create success responses from these nodes
        communication_list list -- create custom responses
        name string -- the key of this call
        """
        place_multinode_call(
            self.__calls,
            name,
            node_labels,
            communication_list,
            action="remote/cluster_start",
        )

    def check_pacemaker_started(
        self,
        pacemaker_started_node_list=(),
        pacemaker_not_started_node_list=(),
        communication_list=None,
        name="http.host.check_pacemaker_started",
    ):
        """
        Create a call for checking pacemaker status on nodes.

        pacemaker_started_node_list list -- list of node names on which
            pacemaker is fully running
        pacemaker_not_started_node_list list -- listof node names on which
            pacemaker is not fully started yet
        communication_list list -- create custom responses
        name string -- the key of this call
        """
        if bool(
            pacemaker_started_node_list or pacemaker_not_started_node_list
        ) == bool(communication_list):
            raise AssertionError(
                "Exactly one of 'pacemaker_started_node_list and/or "
                "pacemaker_not_started_node_list', 'communication_list' must "
                "be specified"
            )
        if not communication_list:
            communication_list = [
                dict(
                    label=node,
                    output=json.dumps(
                        dict(
                            pending=False,
                            online=True,
                        )
                    ),
                )
                for node in pacemaker_started_node_list
            ] + [
                dict(
                    label=node,
                    output=json.dumps(
                        dict(
                            pending=True,
                            online=False,
                        )
                    ),
                )
                for node in pacemaker_not_started_node_list
            ]

        place_communication(
            self.__calls,
            name,
            communication_list,
            action="remote/pacemaker_node_status",
        )

    def get_quorum_status(
        self,
        node_list=None,
        node_labels=None,
        communication_list=None,
        name="http.host.get_quorum_status",
    ):
        output = ""
        if node_list:
            output = outdent(
                """\
            Quorum information
            ------------------
            Date:             Fri Jan 16 13:03:28 2015
            Quorum provider:  corosync_votequorum
            Nodes:            {nodes_num}
            Node ID:          1
            Ring ID:          19860
            Quorate:          Yes\n
            Votequorum information
            ----------------------
            Expected votes:   {nodes_num}
            Highest expected: {nodes_num}
            Total votes:      {nodes_num}
            Quorum:           {quorum_num}
            Flags:            Quorate\n
            Membership information
            ----------------------
                Nodeid      Votes    Qdevice Name
            {nodes}\
            """
            ).format(
                nodes_num=len(node_list),
                quorum_num=(len(node_list) // 2) + 1,
                nodes="".join(
                    [
                        _quorum_status_node_fixture(node_id, node)
                        for node_id, node in enumerate(node_list, 1)
                    ]
                ),
            )

        place_multinode_call(
            self.__calls,
            name,
            node_labels,
            communication_list,
            action="remote/get_quorum_info",
            output=output,
        )


def _quorum_status_node_fixture(node_id, node_name, votes=1, is_local=False):
    _local = " (local)" if is_local else ""
    return (
        f"         {node_id}          {votes}         NR {node_name}{_local}\n"
    )



import os.path
import sys

CURRENT_DIR = os.path.dirname(os.path.abspath(__file__))
DATA_DIR = os.path.join(CURRENT_DIR, "pacemaker_metadata.d")


def write_file_to_stdout(file_name):
    with open(file_name) as file:
        sys.stdout.write(file.read())


def write_local_file_to_stdout(file_name):
    write_file_to_stdout(os.path.join(DATA_DIR, file_name))


def main():
    pcmk_binary = sys.argv[0].rsplit("/", 1)[-1]
    argv = sys.argv[1:]
    if pcmk_binary == "pacemaker-fenced":
        return pacemaker_fenced(argv)
    raise AssertionError()


def pacemaker_fenced(argv):
    if len(argv) != 1:
        raise AssertionError()
    if argv[0] != "metadata":
        raise AssertionError()
    write_local_file_to_stdout("pacemaker_fenced.xml")


if __name__ == "__main__":
    main()



import os
import signal
import socket
from pathlib import Path

from tornado.ioloop import IOLoop
from tornado.locks import Lock
from tornado.web import Application

from pcs import settings
from pcs.daemon import (
    log,
    ruby_pcsd,
    session,
    ssl,
    systemd,
)
from pcs.daemon.app import (
    sinatra_remote,
    sinatra_ui,
    ui,
)
from pcs.daemon.app.common import RedirectHandler
from pcs.daemon.env import prepare_env
from pcs.daemon.http_server import HttpsServerManage


class SignalInfo:
    # pylint: disable=too-few-public-methods
    server_manage = None
    ioloop_started = False


def handle_signal(incoming_signal, frame):
    # pylint: disable=unused-argument
    log.pcsd.warning("Caught signal: %s, shutting down", incoming_signal)
    if SignalInfo.server_manage:
        SignalInfo.server_manage.stop()
    if SignalInfo.ioloop_started:
        IOLoop.current().stop()
    raise SystemExit(0)


def sign_ioloop_started():
    SignalInfo.ioloop_started = True


def config_sync(sync_config_lock: Lock, ruby_pcsd_wrapper: ruby_pcsd.Wrapper):
    async def config_synchronization():
        async with sync_config_lock:
            next_run_time = await ruby_pcsd_wrapper.sync_configs()
        IOLoop.current().call_at(next_run_time, config_synchronization)

    return config_synchronization


def configure_app(
    session_storage: session.Storage,
    ruby_pcsd_wrapper: ruby_pcsd.Wrapper,
    sync_config_lock: Lock,
    public_dir,
    disable_gui=False,
    debug=False,
):
    def make_app(https_server_manage: HttpsServerManage):
        """
        https_server_manage -- allows to control the server (specifically
            reload its SSL certificates). A relevant handler should get this
            object via the method `initialize`.
        """
        routes = sinatra_remote.get_routes(
            ruby_pcsd_wrapper,
            sync_config_lock,
            https_server_manage,
        )

        if not disable_gui:
            routes.extend(
                # old web ui by default
                [(r"/", RedirectHandler, dict(url="/manage"))]
                + [(r"/ui", RedirectHandler, dict(url="/ui/"))]
                + ui.get_routes(
                    url_prefix="/ui/",
                    app_dir=os.path.join(public_dir, "ui"),
                    fallback_page_path=os.path.join(
                        public_dir,
                        "ui_instructions.html",
                    ),
                    session_storage=session_storage,
                )
                + sinatra_ui.get_routes(
                    session_storage, ruby_pcsd_wrapper, public_dir
                )
            )

        return Application(routes, debug=debug)

    return make_app


def main():
    signal.signal(signal.SIGTERM, handle_signal)
    signal.signal(signal.SIGINT, handle_signal)

    Path(settings.pcsd_log_location).touch(mode=0o600, exist_ok=True)
    log.setup(settings.pcsd_log_location)

    env = prepare_env(os.environ, log.pcsd)
    if env.has_errors:
        raise SystemExit(1)

    if env.PCSD_DEBUG:
        log.enable_debug()

    sync_config_lock = Lock()
    ruby_pcsd_wrapper = ruby_pcsd.Wrapper(
        settings.pcsd_ruby_socket,
        debug=env.PCSD_DEBUG,
    )
    make_app = configure_app(
        session.Storage(env.PCSD_SESSION_LIFETIME),
        ruby_pcsd_wrapper,
        sync_config_lock,
        env.PCSD_STATIC_FILES_DIR,
        disable_gui=env.PCSD_DISABLE_GUI,
        debug=env.PCSD_DEV,
    )
    pcsd_ssl = ssl.PcsdSSL(
        server_name=socket.gethostname(),
        cert_location=settings.pcsd_cert_location,
        key_location=settings.pcsd_key_location,
        ssl_options=env.PCSD_SSL_OPTIONS,
        ssl_ciphers=env.PCSD_SSL_CIPHERS,
    )
    try:
        SignalInfo.server_manage = HttpsServerManage(
            make_app,
            port=env.PCSD_PORT,
            bind_addresses=env.PCSD_BIND_ADDR,
            ssl=pcsd_ssl,
        ).start()
    except socket.gaierror as e:
        log.pcsd.error(
            "Unable to bind to specific address(es), exiting: %s ", e
        )
        raise SystemExit(1) from e
    except OSError as e:
        log.pcsd.error("Unable to start pcsd daemon, exiting: %s ", e)
        raise SystemExit(1) from e
    except ssl.SSLCertKeyException as e:
        for error in e.args:
            log.pcsd.error(error)
        log.pcsd.error("Invalid SSL certificate and/or key, exiting")
        raise SystemExit(1) from e

    ioloop = IOLoop.current()
    ioloop.add_callback(sign_ioloop_started)
    if systemd.is_systemd() and env.NOTIFY_SOCKET:
        ioloop.add_callback(systemd.notify, env.NOTIFY_SOCKET)
    ioloop.add_callback(config_sync(sync_config_lock, ruby_pcsd_wrapper))
    ioloop.start()



import os.path

BIN_MOCK_DIR = os.path.dirname(os.path.abspath(__file__))

CRM_RESOURCE_BIN = os.path.abspath(
    os.path.join(BIN_MOCK_DIR, "pcmk/crm_resource")
)
PACEMAKER_FENCED_BIN = os.path.abspath(
    os.path.join(BIN_MOCK_DIR, "pcmk/pacemaker-fenced")
)

MOCK_SETTINGS = {
    "crm_resource_binary": CRM_RESOURCE_BIN,
    "pacemaker_fenced": PACEMAKER_FENCED_BIN,
}


def get_mock_settings(*required_settings):
    return {
        key: value
        for key, value in MOCK_SETTINGS.items()
        if key in required_settings
    }



from pcs import (
    acl,
    usage,
)
from pcs.cli.common.routing import create_router

acl_cmd = create_router(
    {
        "help": lambda lib, argv, modifiers: usage.acl(argv),
        # TODO remove, deprecated command
        # replaced with 'config'
        "show": acl.show_acl_config,
        "config": acl.acl_config,
        "enable": acl.acl_enable,
        "disable": acl.acl_disable,
        "role": create_router(
            {
                "create": acl.role_create,
                "delete": acl.role_delete,
                "remove": acl.role_delete,
                "assign": acl.role_assign,
                "unassign": acl.role_unassign,
            },
            ["acl", "role"],
        ),
        "user": create_router(
            {
                "create": acl.user_create,
                "delete": acl.user_delete,
                "remove": acl.user_delete,
            },
            ["acl", "user"],
        ),
        "group": create_router(
            {
                "create": acl.group_create,
                "delete": acl.group_delete,
                "remove": acl.group_delete,
            },
            ["acl", "group"],
        ),
        "permission": create_router(
            {
                "add": acl.permission_add,
                "delete": acl.run_permission_delete,
                "remove": acl.run_permission_delete,
            },
            ["acl", "permission"],
        ),
    },
    ["acl"],
    default_cmd="config",
)



from unittest import TestCase

from lxml import etree

from pcs.common.tools import Version
from pcs.lib.cib import rule
from pcs.lib.cib.rule.expression_part import (
    BOOL_AND,
    BOOL_OR,
    DATE_OP_GT,
    DATE_OP_LT,
    NODE_ATTR_OP_DEFINED,
    NODE_ATTR_OP_EQ,
    NODE_ATTR_OP_GT,
    NODE_ATTR_OP_GTE,
    NODE_ATTR_OP_LT,
    NODE_ATTR_OP_LTE,
    NODE_ATTR_OP_NE,
    NODE_ATTR_OP_NOT_DEFINED,
    NODE_ATTR_TYPE_INTEGER,
    NODE_ATTR_TYPE_NUMBER,
    NODE_ATTR_TYPE_STRING,
    NODE_ATTR_TYPE_VERSION,
    BoolExpr,
    DateInRangeExpr,
    DatespecExpr,
    DateUnaryExpr,
    NodeAttrExpr,
    OpExpr,
    RscExpr,
)
from pcs.lib.cib.tools import IdProvider

from pcs_test.tools.assertions import assert_xml_equal
from pcs_test.tools.xml import etree_to_str


class Base(TestCase):
    @staticmethod
    def assert_cib(tree, expected_xml, schema_version=None):
        if schema_version is None:
            schema_version = Version(3, 5, 0)
        xml = etree.fromstring('<root id="X"/>')
        rule.rule_to_cib(xml, IdProvider(xml), schema_version, tree)
        assert_xml_equal(
            '<root id="X">' + expected_xml + "</root>", etree_to_str(xml)
        )


class SimpleBool(Base):
    def test_no_children(self):
        self.assert_cib(
            BoolExpr(BOOL_AND, []),
            """
                <rule id="X-rule" boolean-op="and" score="INFINITY" />
            """,
        )

    def test_one_child(self):
        self.assert_cib(
            BoolExpr(BOOL_AND, [OpExpr("start", None)]),
            """
                <rule id="X-rule" boolean-op="and" score="INFINITY">
                    <op_expression id="X-rule-op-start" name="start" />
                </rule>
            """,
        )

    def test_two_children(self):
        operators = [
            (BOOL_OR, "or"),
            (BOOL_AND, "and"),
        ]
        for op_in, op_out in operators:
            with self.subTest(op_in=op_in, op_out=op_out):
                self.assert_cib(
                    BoolExpr(
                        op_in,
                        [
                            OpExpr("start", None),
                            RscExpr("systemd", None, "pcsd"),
                        ],
                    ),
                    f"""
                        <rule id="X-rule" boolean-op="{op_out}" score="INFINITY">
                            <op_expression id="X-rule-op-start" name="start" />
                            <rsc_expression id="X-rule-rsc-systemd-pcsd"
                                class="systemd" type="pcsd"
                            />
                        </rule>
                    """,
                )


class SimpleNodeAttr(Base):
    def test_defined(self):
        self.assert_cib(
            NodeAttrExpr(NODE_ATTR_OP_DEFINED, "pingd", None, None),
            """
                <expression attribute="pingd" id="X-expr" operation="defined" />
            """,
        )

    def test_not_defined(self):
        self.assert_cib(
            NodeAttrExpr(NODE_ATTR_OP_NOT_DEFINED, "pingd", None, None),
            """
                <expression attribute="pingd" id="X-expr" operation="not_defined" />
            """,
        )

    def test_eq(self):
        self.assert_cib(
            NodeAttrExpr(NODE_ATTR_OP_EQ, "#uname", "node1", None),
            """
                <expression attribute="#uname" id="X-expr" operation="eq"
                    value="node1"
                />
            """,
        )

    def test_ne(self):
        self.assert_cib(
            NodeAttrExpr(NODE_ATTR_OP_NE, "#uname", "node1", None),
            """
                <expression attribute="#uname" id="X-expr" operation="ne"
                    value="node1"
                />
            """,
        )

    def test_gt(self):
        self.assert_cib(
            NodeAttrExpr(NODE_ATTR_OP_GT, "#uname", "node1", None),
            """
                <expression attribute="#uname" id="X-expr" operation="gt"
                    value="node1"
                />
            """,
        )

    def test_gte(self):
        self.assert_cib(
            NodeAttrExpr(NODE_ATTR_OP_GTE, "#uname", "node1", None),
            """
                <expression attribute="#uname" id="X-expr" operation="gte"
                    value="node1"
                />
            """,
        )

    def test_lt(self):
        self.assert_cib(
            NodeAttrExpr(NODE_ATTR_OP_LT, "#uname", "node1", None),
            """
                <expression attribute="#uname" id="X-expr" operation="lt"
                    value="node1"
                />
            """,
        )

    def test_lte(self):
        self.assert_cib(
            NodeAttrExpr(NODE_ATTR_OP_LTE, "#uname", "node1", None),
            """
                <expression attribute="#uname" id="X-expr" operation="lte"
                    value="node1"
                />
            """,
        )

    def test_type_integer(self):
        self.assert_cib(
            NodeAttrExpr(
                NODE_ATTR_OP_EQ, "#uname", "12345", NODE_ATTR_TYPE_INTEGER
            ),
            """
                <expression attribute="#uname" id="X-expr" operation="eq"
                    type="integer" value="12345"
                />
            """,
        )

    def test_type_integer_old_schema(self):
        self.assert_cib(
            NodeAttrExpr(
                NODE_ATTR_OP_EQ, "#uname", "12345", NODE_ATTR_TYPE_INTEGER
            ),
            """
                <expression attribute="#uname" id="X-expr" operation="eq"
                    type="number" value="12345"
                />
            """,
            Version(3, 4, 0),
        )

    def test_type_number(self):
        self.assert_cib(
            NodeAttrExpr(
                NODE_ATTR_OP_EQ, "#uname", "12345", NODE_ATTR_TYPE_NUMBER
            ),
            """
                <expression attribute="#uname" id="X-expr" operation="eq"
                    type="number" value="12345"
                />
            """,
        )

    def test_type_number_old_schema(self):
        self.assert_cib(
            NodeAttrExpr(
                NODE_ATTR_OP_EQ, "#uname", "12345", NODE_ATTR_TYPE_NUMBER
            ),
            """
                <expression attribute="#uname" id="X-expr" operation="eq"
                    type="number" value="12345"
                />
            """,
            Version(3, 4, 0),
        )

    def test_type_string(self):
        self.assert_cib(
            NodeAttrExpr(
                NODE_ATTR_OP_EQ, "#uname", "node1", NODE_ATTR_TYPE_STRING
            ),
            """
                <expression attribute="#uname" id="X-expr" operation="eq"
                    type="string" value="node1"
                />
            """,
        )

    def test_type_version(self):
        self.assert_cib(
            NodeAttrExpr(
                NODE_ATTR_OP_EQ, "#uname", "1.2.3", NODE_ATTR_TYPE_VERSION
            ),
            """
                <expression attribute="#uname" id="X-expr" operation="eq"
                    type="version" value="1.2.3"
                />
            """,
        )


class SimpleDatespec(Base):
    def test_1(self):
        self.assert_cib(
            DatespecExpr([("hours", "1")]),
            """
                <date_expression id="X-expr" operation="date_spec">
                    <date_spec id="X-expr-datespec" hours="1" />
                </date_expression>
            """,
        )

    def test_2(self):
        self.assert_cib(
            DatespecExpr(
                [("hours", "1-14"), ("monthdays", "20-30"), ("months", "1")]
            ),
            """
                <date_expression id="X-expr" operation="date_spec">
                    <date_spec id="X-expr-datespec"
                        hours="1-14" monthdays="20-30" months="1"
                    />
                </date_expression>
            """,
        )


class SimpleDate(Base):
    def test_gt(self):
        self.assert_cib(
            DateUnaryExpr(DATE_OP_GT, "2014-06-26"),
            """
                <date_expression id="X-expr" operation="gt" start="2014-06-26" />
            """,
        )

    def test_lt(self):
        self.assert_cib(
            DateUnaryExpr(DATE_OP_LT, "2014-06-26"),
            """
                <date_expression id="X-expr" operation="lt" end="2014-06-26" />
            """,
        )

    def test_inrange_start_end(self):
        self.assert_cib(
            DateInRangeExpr("2014-06-26", "2014-07-26", None),
            """
                <date_expression id="X-expr"
                    operation="in_range" start="2014-06-26" end="2014-07-26"
                />
            """,
        )

    def test_inrange_end(self):
        self.assert_cib(
            DateInRangeExpr(None, "2014-07-26", None),
            """
                <date_expression id="X-expr"
                    operation="in_range" end="2014-07-26"
                />
            """,
        )

    def test_inrange_start_duration(self):
        self.assert_cib(
            DateInRangeExpr("2014-06-26", None, [("years", "1")]),
            """
                <date_expression id="X-expr" operation="in_range" start="2014-06-26">
                    <duration id="X-expr-duration" years="1"/>
                </date_expression>
            """,
        )


class SimpleOp(Base):
    def test_minimal(self):
        self.assert_cib(
            OpExpr("start", None),
            """
                <op_expression id="X-op-start" name="start" />
            """,
        )

    def test_interval(self):
        self.assert_cib(
            OpExpr("monitor", "2min"),
            """
                <op_expression id="X-op-monitor" name="monitor"
                    interval="2min"
                />
            """,
        )


class SimpleRsc(Base):
    def test_class(self):
        self.assert_cib(
            RscExpr("ocf", None, None),
            """
                <rsc_expression id="X-rsc-ocf" class="ocf" />
            """,
        )

    def test_provider(self):
        self.assert_cib(
            RscExpr(None, "pacemaker", None),
            """
                <rsc_expression id="X-rsc-pacemaker" provider="pacemaker" />
            """,
        )

    def type(self):
        self.assert_cib(
            RscExpr(None, None, "Dummy"),
            """
                <rsc_expression id="X-rsc-Dummy" type="Dummy" />
            """,
        )

    def test_provider_type(self):
        self.assert_cib(
            RscExpr(None, "pacemaker", "Dummy"),
            """
                <rsc_expression id="X-rsc-pacemaker-Dummy"
                    provider="pacemaker" type="Dummy"
                />
            """,
        )

    def test_class_provider(self):
        self.assert_cib(
            RscExpr("ocf", "pacemaker", None),
            """
                <rsc_expression id="X-rsc-ocf-pacemaker"
                    class="ocf" provider="pacemaker"
                />
            """,
        )

    def test_class_type(self):
        self.assert_cib(
            RscExpr("systemd", None, "pcsd"),
            """
                <rsc_expression id="X-rsc-systemd-pcsd"
                    class="systemd" type="pcsd"
                />
            """,
        )

    def test_class_provider_type(self):
        self.assert_cib(
            RscExpr("ocf", "pacemaker", "Dummy"),
            """
                <rsc_expression id="X-rsc-ocf-pacemaker-Dummy"
                    class="ocf" provider="pacemaker" type="Dummy"
                />
            """,
        )


class Complex(Base):
    def test_expr_1(self):
        self.assert_cib(
            BoolExpr(
                BOOL_AND,
                [
                    BoolExpr(
                        BOOL_OR,
                        [
                            RscExpr("ocf", "pacemaker", "Dummy"),
                            OpExpr("start", None),
                            RscExpr("systemd", None, "pcsd"),
                            RscExpr("ocf", "heartbeat", "Dummy"),
                        ],
                    ),
                    BoolExpr(
                        BOOL_OR,
                        [
                            OpExpr("monitor", "30s"),
                            RscExpr("ocf", "pacemaker", "Dummy"),
                            OpExpr("start", None),
                            OpExpr("monitor", "2min"),
                        ],
                    ),
                ],
            ),
            """
                <rule id="X-rule" boolean-op="and" score="INFINITY">
                  <rule id="X-rule-rule" boolean-op="or" score="0">
                    <rsc_expression id="X-rule-rule-rsc-ocf-pacemaker-Dummy"
                        class="ocf" provider="pacemaker" type="Dummy"
                    />
                    <op_expression id="X-rule-rule-op-start" name="start" />
                    <rsc_expression id="X-rule-rule-rsc-systemd-pcsd"
                        class="systemd" type="pcsd"
                    />
                    <rsc_expression id="X-rule-rule-rsc-ocf-heartbeat-Dummy"
                        class="ocf" provider="heartbeat" type="Dummy"
                    />
                  </rule>
                  <rule id="X-rule-rule-1" boolean-op="or" score="0">
                    <op_expression id="X-rule-rule-1-op-monitor"
                        name="monitor" interval="30s"
                    />
                    <rsc_expression id="X-rule-rule-1-rsc-ocf-pacemaker-Dummy"
                        class="ocf" provider="pacemaker" type="Dummy"
                    />
                    <op_expression id="X-rule-rule-1-op-start" name="start" />
                    <op_expression id="X-rule-rule-1-op-monitor-1"
                        name="monitor" interval="2min"
                    />
                  </rule>
                </rule>
            """,
        )



from unittest import TestCase

from lxml import etree

from pcs_test.tools.cib import get_assert_pcs_effect_mixin
from pcs_test.tools.misc import get_test_resource as rc
from pcs_test.tools.misc import (
    get_tmp_file,
    write_file_to_tmpfile,
)
from pcs_test.tools.pcs_runner import PcsRunner


class ManageUnmanage(
    TestCase,
    get_assert_pcs_effect_mixin(
        lambda cib: etree.tostring(
            # pylint:disable=undefined-variable
            etree.parse(cib).findall(".//resources")[0]
        )
    ),
):
    empty_cib = rc("cib-empty.xml")

    @staticmethod
    def fixture_cib_unmanaged_a(add_empty_meta_b=False):
        empty_meta_b = '<meta_attributes id="B-meta_attributes" />'
        return """
            <resources>
                <primitive class="ocf" id="A" provider="heartbeat" type="Dummy">
                    <meta_attributes id="A-meta_attributes">
                        <nvpair id="A-meta_attributes-is-managed"
                            name="is-managed" value="false"
                        />
                    </meta_attributes>
                    <operations>
                        <op id="A-monitor-interval-10s" interval="10s"
                            name="monitor" timeout="20s"
                        />
                    </operations>
                </primitive>
                <primitive class="ocf" id="B" provider="heartbeat" type="Dummy">
                    {empty_meta_b}<operations>
                        <op id="B-monitor-interval-10s" interval="10s"
                            name="monitor" timeout="20s"
                        />
                    </operations>
                </primitive>
            </resources>
        """.format(
            empty_meta_b=(empty_meta_b if add_empty_meta_b else "")
        )

    def setUp(self):
        self.temp_cib = get_tmp_file("tier1_cib_resource_manage_unmanage")
        write_file_to_tmpfile(self.empty_cib, self.temp_cib)
        self.pcs_runner = PcsRunner(self.temp_cib.name)

    def tearDown(self):
        self.temp_cib.close()

    def fixture_resource(self, name, managed=True, with_monitors=False):
        self.assert_pcs_success(
            [
                "resource",
                "create",
                name,
                "ocf:heartbeat:Dummy",
                "--no-default-ops",
            ]
        )
        if not managed:
            cmd = ["resource", "unmanage", name]
            if with_monitors:
                cmd.append("--monitor")
            self.assert_pcs_success(cmd)

    def fixture_tag(self, name, ids):
        self.assert_pcs_success(["tag", "create", name] + ids)

    def test_unmanage_none(self):
        self.assert_pcs_fail(
            "resource unmanage".split(),
            "Error: You must specify resource(s) to unmanage\n",
        )

    def test_manage_none(self):
        self.assert_pcs_fail(
            "resource manage".split(),
            "Error: You must specify resource(s) to manage\n",
        )

    def test_unmanage_one(self):
        self.fixture_resource("A")
        self.fixture_resource("B")
        self.assert_effect(
            "resource unmanage A".split(), self.fixture_cib_unmanaged_a()
        )

    def test_manage_one(self):
        self.fixture_resource("A", managed=False)
        self.fixture_resource("B", managed=False)
        self.assert_effect(
            "resource manage B".split(),
            self.fixture_cib_unmanaged_a(add_empty_meta_b=True),
        )

    def test_unmanage_monitor(self):
        self.fixture_resource("A")
        self.assert_effect(
            "resource unmanage A --monitor".split(),
            """
            <resources>
                <primitive class="ocf" id="A" provider="heartbeat" type="Dummy">
                    <meta_attributes id="A-meta_attributes">
                        <nvpair id="A-meta_attributes-is-managed"
                            name="is-managed" value="false"
                        />
                    </meta_attributes>
                    <operations>
                        <op id="A-monitor-interval-10s" interval="10s"
                            name="monitor" timeout="20s" enabled="false"
                        />
                    </operations>
                </primitive>
            </resources>
            """,
        )

    def test_unmanage_monitor_enabled(self):
        self.fixture_resource("A")
        self.assert_effect(
            "resource unmanage A".split(),
            """
            <resources>
                <primitive class="ocf" id="A" provider="heartbeat" type="Dummy">
                    <meta_attributes id="A-meta_attributes">
                        <nvpair id="A-meta_attributes-is-managed"
                            name="is-managed" value="false"
                        />
                    </meta_attributes>
                    <operations>
                        <op id="A-monitor-interval-10s" interval="10s"
                            name="monitor" timeout="20s"
                        />
                    </operations>
                </primitive>
            </resources>
            """,
        )

    def test_manage_monitor(self):
        self.fixture_resource("A", managed=True, with_monitors=True)
        self.assert_effect(
            "resource manage A --monitor".split(),
            """
            <resources>
                <primitive class="ocf" id="A" provider="heartbeat" type="Dummy">
                    <operations>
                        <op id="A-monitor-interval-10s" interval="10s"
                            name="monitor" timeout="20s"
                        />
                    </operations>
                </primitive>
            </resources>
            """,
        )

    def test_manage_monitor_disabled(self):
        self.fixture_resource("A", managed=False, with_monitors=True)
        self.assert_effect(
            "resource manage A".split(),
            """
            <resources>
                <primitive class="ocf" id="A" provider="heartbeat" type="Dummy">
                    <meta_attributes id="A-meta_attributes" />
                    <operations>
                        <op id="A-monitor-interval-10s" interval="10s"
                            name="monitor" timeout="20s" enabled="false"
                        />
                    </operations>
                </primitive>
            </resources>
            """,
            "Warning: Resource 'A' has no enabled monitor operations."
            " Re-run with '--monitor' to enable them.\n",
        )

    def test_unmanage_more(self):
        self.fixture_resource("A")
        self.fixture_resource("B")
        self.fixture_tag("TA", ["A"])
        self.assert_effect(
            "resource unmanage TA B".split(),
            """
            <resources>
                <primitive class="ocf" id="A" provider="heartbeat" type="Dummy">
                    <meta_attributes id="A-meta_attributes">
                        <nvpair id="A-meta_attributes-is-managed"
                            name="is-managed" value="false"
                        />
                    </meta_attributes>
                    <operations>
                        <op id="A-monitor-interval-10s" interval="10s"
                            name="monitor" timeout="20s"
                        />
                    </operations>
                </primitive>
                <primitive class="ocf" id="B" provider="heartbeat" type="Dummy">
                    <meta_attributes id="B-meta_attributes">
                        <nvpair id="B-meta_attributes-is-managed"
                            name="is-managed" value="false"
                        />
                    </meta_attributes>
                    <operations>
                        <op id="B-monitor-interval-10s" interval="10s"
                            name="monitor" timeout="20s"
                        />
                    </operations>
                </primitive>
            </resources>
            """,
        )

    def test_manage_more(self):
        self.fixture_resource("A", managed=False)
        self.fixture_resource("B", managed=False)
        self.fixture_tag("TA", ["A"])
        self.assert_effect(
            "resource manage TA B".split(),
            """
            <resources>
                <primitive class="ocf" id="A" provider="heartbeat" type="Dummy">
                    <meta_attributes id="A-meta_attributes" />
                    <operations>
                        <op id="A-monitor-interval-10s" interval="10s"
                            name="monitor" timeout="20s"
                        />
                    </operations>
                </primitive>
                <primitive class="ocf" id="B" provider="heartbeat" type="Dummy">
                    <meta_attributes id="B-meta_attributes" />
                    <operations>
                        <op id="B-monitor-interval-10s" interval="10s"
                            name="monitor" timeout="20s"
                        />
                    </operations>
                </primitive>
            </resources>
            """,
        )

    def test_unmanage_nonexistent(self):
        self.fixture_resource("A")

        self.assert_pcs_fail(
            "resource unmanage A B".split(),
            (
                "Error: bundle/clone/group/resource/tag 'B' does not exist\n"
                "Error: Errors have occurred, therefore pcs is unable to continue\n"
            ),
        )
        self.assert_resources_xml_in_cib(
            """
            <resources>
                <primitive class="ocf" id="A" provider="heartbeat" type="Dummy">
                    <operations>
                        <op id="A-monitor-interval-10s" interval="10s"
                            name="monitor" timeout="20s"
                        />
                    </operations>
                </primitive>
            </resources>
            """
        )

    def test_manage_nonexistent(self):
        self.fixture_resource("A", managed=False)

        self.assert_pcs_fail(
            "resource manage A B".split(),
            (
                "Error: bundle/clone/group/resource/tag 'B' does not exist\n"
                "Error: Errors have occurred, therefore pcs is unable to continue\n"
            ),
        )
        self.assert_resources_xml_in_cib(
            """
            <resources>
                <primitive class="ocf" id="A" provider="heartbeat" type="Dummy">
                    <meta_attributes id="A-meta_attributes">
                        <nvpair id="A-meta_attributes-is-managed"
                            name="is-managed" value="false"
                        />
                    </meta_attributes>
                    <operations>
                        <op id="A-monitor-interval-10s" interval="10s"
                            name="monitor" timeout="20s"
                        />
                    </operations>
                </primitive>
            </resources>
            """
        )



from pcs import settings

from pcs_test.tools.command_env.mock_runner import Call as RunnerCall


class BoothShortcuts:
    def __init__(self, calls):
        self.__calls = calls

    def status_daemon(
        self,
        instance_name=None,
        stdout="",
        stderr="",
        returncode=0,
        name="runner.booth.status_daemon",
        instead=None,
        before=None,
    ):
        """
        Create a call for getting the booth daemon status

        string instance_name -- booth instance name
        string stdout -- stdout of the booth command
        string stderr -- stderr of the booth command
        int returncode -- returncode of the booth command
        string name -- the key of the call
        string before -- the key of a call before which this call is to be
            placed
        string instead -- the key of a call instead of which this new call is to
            be placed
        """
        cmd = [settings.booth_binary, "status"]
        if instance_name:
            cmd.extend(["-c", instance_name])
        self.__calls.place(
            name,
            RunnerCall(
                cmd,
                stdout=stdout,
                stderr=stderr,
                returncode=returncode,
            ),
            before=before,
            instead=instead,
        )

    def status_peers(
        self,
        instance_name=None,
        stdout="",
        stderr="",
        returncode=0,
        name="runner.booth.status_peers",
        instead=None,
        before=None,
    ):
        """
        Create a call for getting the booth peers status

        string instance_name -- booth instance name
        string stdout -- stdout of the booth command
        string stderr -- stderr of the booth command
        int returncode -- returncode of the booth command
        string name -- the key of the call
        string before -- the key of a call before which this call is to be
            placed
        string instead -- the key of a call instead of which this new call is to
            be placed
        """
        cmd = [settings.booth_binary, "peers"]
        if instance_name:
            cmd.extend(["-c", instance_name])
        self.__calls.place(
            name,
            RunnerCall(
                cmd,
                stdout=stdout,
                stderr=stderr,
                returncode=returncode,
            ),
            before=before,
            instead=instead,
        )

    def status_tickets(
        self,
        instance_name=None,
        stdout="",
        stderr="",
        returncode=0,
        name="runner.booth.status_tickets",
        instead=None,
        before=None,
    ):
        """
        Create a call for getting the booth tickets status

        string instance_name -- booth instance name
        string stdout -- stdout of the booth command
        string stderr -- stderr of the booth command
        int returncode -- returncode of the booth command
        string name -- the key of the call
        string before -- the key of a call before which this call is to be
            placed
        string instead -- the key of a call instead of which this new call is to
            be placed
        """
        cmd = [settings.booth_binary, "list"]
        if instance_name:
            cmd.extend(["-c", instance_name])
        self.__calls.place(
            name,
            RunnerCall(
                cmd,
                stdout=stdout,
                stderr=stderr,
                returncode=returncode,
            ),
            before=before,
            instead=instead,
        )

    def ticket_grant(
        self,
        ticket_name,
        site_ip,
        stdout="",
        stderr="",
        returncode=0,
        name="runner.booth.ticket_grant",
        instead=None,
        before=None,
    ):
        # pylint: disable=too-many-arguments
        """
        Create a call for granting a ticket

        string ticket_name -- the name of the ticket to be granted
        string site_ip -- an IP address of a site the ticket is being granted to
        string stdout -- stdout of the booth grant command
        string stderr -- stderr of the booth grant command
        int returncode -- returncode of the booth grant command
        string name -- the key of the call
        string before -- the key of a call before which this call is to be
            placed
        string instead -- the key of a call instead of which this new call is to
            be placed
        """
        self.__calls.place(
            name,
            RunnerCall(
                [settings.booth_binary, "grant", "-s", site_ip, ticket_name],
                stdout=stdout,
                stderr=stderr,
                returncode=returncode,
            ),
            before=before,
            instead=instead,
        )

    def ticket_revoke(
        self,
        ticket_name,
        site_ip,
        stdout="",
        stderr="",
        returncode=0,
        name="runner.booth.ticket_revoke",
        instead=None,
        before=None,
    ):
        # pylint: disable=too-many-arguments
        """
        Create a call for revoking a ticket

        string ticket_name -- the name of the ticket to be revoked
        string site_ip -- an IP address of a site the ticket is being revoked to
        string stdout -- stdout of the booth revoke command
        string stderr -- stderr of the booth revoke command
        int returncode -- returncode of the booth revoke command
        string name -- the key of the call
        string before -- the key of a call before which this call is to be
            placed
        string instead -- the key of a call instead of which this new call is to
            be placed
        """
        self.__calls.place(
            name,
            RunnerCall(
                [settings.booth_binary, "revoke", "-s", site_ip, ticket_name],
                stdout=stdout,
                stderr=stderr,
                returncode=returncode,
            ),
            before=before,
            instead=instead,
        )



# This module is intended to test just the new Communicator/MultiringCommunicator
# classes which are using curllib

# pylint: disable=wrong-import-position, no-member, protected-access

import logging
import os.path
import pprint
import sys

PACKAGE_DIR = os.path.dirname(os.path.abspath(__file__))
sys.path.insert(0, PACKAGE_DIR)

from pcs import utils
from pcs.common.host import Destination
from pcs.common.node_communicator import (
    Request,
    RequestData,
    RequestTarget,
)

logger_handler = logging.StreamHandler()
logger_handler.setFormatter(logging.Formatter("%(levelname)s: %(message)s"))
logger = logging.getLogger("pcs")
logger.setLevel(logging.DEBUG)
logger.addHandler(logger_handler)

global_target = RequestTarget(
    "TestServer",
    dest_list=[
        Destination("httpbin.org2", 433),
        Destination("httpbin.org", 443),
    ],
)

pprint.pprint(global_target)


def get_request(timeout):
    return Request(global_target, RequestData("delay/{0}".format(timeout)))


lib_env = utils.get_lib_env()
# utils.pcs_options["--debug"] = True
request_list = [get_request((i + 1) * 2) for i in range(6)]
factory = lib_env.get_node_communicator_factory()
factory._request_timeout = 10
communicator = factory.get_multiring_communicator()
# communicator.add_requests([get_request(10)])
# response = list(communicator.start_loop())[0]
# pprint.pprint(response.to_report_item())
communicator.add_requests(request_list)
for response in communicator.start_loop():
    # print(80 * "-")
    # print(response.request.url)
    # print(response.data)
    # print(80 * "-")
    if response.request == request_list[2]:
        r = get_request(5)
        request_list.append(r)
        communicator.add_requests([r])
    if response.request == request_list[5]:
        r = get_request(10)
        request_list.append(r)
        communicator.add_requests([r])
    if len(request_list) == 8 and response.request == request_list[7]:
        r = get_request(15)
        communicator.add_requests([r])



from typing import (
    Any,
    Optional,
)

from pcs.common import (
    file_type_codes,
    reports,
)


class ParserErrorException(Exception):
    pass


class ParserInterface:
    @staticmethod
    def parse(raw_file_data: bytes) -> Any:
        raise NotImplementedError()

    @staticmethod
    def exception_to_report_list(
        exception: ParserErrorException,
        file_type_code: file_type_codes.FileTypeCode,
        file_path: Optional[str],
        force_code: Optional[reports.types.ForceCode],
        is_forced_or_warning: bool,
    ) -> reports.ReportItemList:
        raise NotImplementedError()


class ExporterInterface:
    @staticmethod
    def export(config_structure: Any) -> bytes:
        """
        Export config structure to raw file data (bytes)

        config_structure -- parsed config file
        """
        raise NotImplementedError()


class FacadeInterface:
    # Facades should also implement a 'create' classmethod which creates a new
    # facade with a minimalistic config in it. This method for sure has
    # different interface in each class (depending on which config's facade it
    # is). The create method is not used by the files framework (there is no
    # need and also due to mentioned interface differences). Therefore the
    # create method is not defined here in the interface.
    def __init__(self, parsed_config: Any):
        """
        Create a facade around a parsed config file

        parsed_config -- parsed config file
        """
        self._config = parsed_config

    @property
    def config(self) -> Any:
        """
        Export a parsed config file
        """
        return self._config



from http.cookies import Morsel

from pcs.daemon.auth import (
    authorize_user,
    check_user_groups,
)
from pcs.daemon.session import Storage

# monkeypatch Python 3.6 to add support for samesite attribute
Morsel._reserved["samesite"] = "SameSite"  # pylint: disable=protected-access


PCSD_SESSION = "pcsd.sid"


class Mixin:
    """
    Mixin for tornado.web.RequestHandler
    """

    __session = None
    __cookie_options = {
        "secure": True,
        "httponly": True,
        # rhbz#2097393
        # Prevent a cookie to be sent on cross-site requests, allow it to be
        # sent when navigating to pcs web UI from an external site.
        "samesite": "Lax",
    }

    def initialize(self, session_storage: Storage):
        self.__storage = session_storage

    async def init_session(self):
        self.__session = self.__storage.provide(self.__sid_from_client)
        if self.__session.is_authenticated:
            self.__refresh_auth(
                await check_user_groups(self.__session.username),
                ajax_id=self.__session.ajax_id,
            )

    async def session_auth_user(self, username, password, sign_rejection=True):
        """
        Make user authorization and refresh storage.

        bool sing_rejection -- flag according to which will be decided whether
            to manipulate with session in storage in case of failed
            authorization. It allows not to touch session for ajax calls when
            authorization fails. It keeps previous behavior and should be
            reviewed.
        """
        # initialize session since it should be used without `init_session`
        self.__session = self.__storage.provide(self.__sid_from_client)
        self.__refresh_auth(
            await authorize_user(username, password),
            sign_rejection=sign_rejection,
        )

    @property
    def session(self):
        if self.__session is None:
            raise Exception(
                "Session is not set in session mixin. "
                "Session probably has not been initialized in request handler."
            )
        return self.__session

    def prepare(self):
        """
        Expired sessions are removed before each request that uses sessions (it
        means before each request that is handled by descendant of this mixin).
        """
        self.__storage.drop_expired()

    def session_logout(self):
        if self.__session is not None:
            self.__storage.destroy(self.__session.sid)
        elif self.__sid_from_client is not None:
            self.__storage.destroy(self.__sid_from_client)
        self.__session = self.__storage.provide()

    def sid_to_cookies(self):
        """
        Write the session id into a response cookie.
        """
        self.set_cookie(PCSD_SESSION, self.session.sid, **self.__cookie_options)

    def put_request_cookies_sid_to_response_cookies_sid(self):
        """
        If sid came in the request cookies put it into response cookies. But do
        not start new one.
        """
        # TODO this method should exist temporarily (for sinatra compatibility)
        # pylint: disable=invalid-name
        if self.__sid_from_client is not None:
            self.set_cookie(
                PCSD_SESSION, self.__sid_from_client, **self.__cookie_options
            )

    def was_sid_in_request_cookies(self):
        return self.__sid_from_client is not None

    @property
    def __sid_from_client(self):
        return self.get_cookie(PCSD_SESSION, default=None)

    def __refresh_auth(self, user_auth_info, sign_rejection=True, ajax_id=None):
        if user_auth_info.is_authorized:
            self.__session = self.__storage.login(
                self.__session.sid,
                user_auth_info.name,
                user_auth_info.groups,
                ajax_id,
            )
            self.sid_to_cookies()
        elif sign_rejection:
            self.__session = self.__storage.rejected_user(
                self.__session.sid,
                user_auth_info.name,
            )
            self.sid_to_cookies()



import json
from urllib.parse import parse_qs

from pcs import settings
from pcs.common import pcs_pycurl as pycurl
from pcs.common.host import Destination
from pcs.common.node_communicator import (
    Request,
    RequestData,
    RequestTarget,
    Response,
)

from pcs_test.tools.custom_mock import MockCurlSimple

# pylint: disable=too-many-arguments, protected-access

CALL_TYPE_HTTP_ADD_REQUESTS = "CALL_TYPE_HTTP_ADD_REQUESTS"
CALL_TYPE_HTTP_START_LOOP = "CALL_TYPE_HTTP_START_LOOP"


def log_request(request):
    label_data = [
        ("action", request.action),
        ("label", request.target.label),
        ("data", parse_qs(request.data)),
    ]

    if request.target.dest_list != [
        Destination(request.target.label, settings.pcsd_default_port)
    ]:
        label_data.append(("dest_list", request.target.dest_list))

    return "  ".join(
        ["{0}:'{1}'".format(key, value) for key, value in label_data]
    )


def log_response(response, indent=0):
    label_data = [
        ("action", response.request.action),
        ("label", response.request.target.label),
    ]

    if response.request.target.dest_list != [
        Destination(response.request.target.label, settings.pcsd_default_port),
    ]:
        label_data.append(("dest_list", response.request.target.dest_list))

    label_data.append(("was_connected", response.was_connected))

    if response.was_connected:
        label_data.append(("response_code", response.response_code))
    else:
        label_data.extend(
            [
                ("errno", response.errno),
                ("error_msg", response.error_msg),
            ]
        )

    label_data.append(("data", parse_qs(response.request.data)))

    return "{0}{1}".format(
        "  " * indent,
        "  ".join(
            ["{0}:'{1}'".format(key, value) for key, value in label_data]
        ),
    )


def different_request_lists(expected_request_list, request_list):
    return AssertionError(
        (
            "Method add_request of NodeCommunicator expected"
            " request_list:\n  * {0}\nbut got: \n  * {1}"
        ).format(
            "\n  * ".join(log_request(r) for r in expected_request_list),
            "\n  * ".join(log_request(r) for r in request_list),
        )
    )


def bad_request_list_content(errors):
    return AssertionError(
        (
            "NodeCommunicator.add_request got different requests than "
            "expected:{0}"
        ).format(
            "".join(
                [
                    "\n  call index {call_index}:{call_details}".format(
                        call_index=call_index,
                        call_details="".join(
                            [
                                "\n    mismatch in {option_name}:"
                                "\n      expected: {expected_value}"
                                "\n      real:     {real_value}".format(
                                    option_name=option_name,
                                    expected_value=pair[0],
                                    real_value=pair[1],
                                )
                                for option_name, pair in value.items()
                            ]
                        ),
                    )
                    for call_index, value in errors.items()
                ]
            )
        )
    )


def _communication_to_response(
    label,
    dest_list,
    action,
    param_list,
    response_code,
    output,
    debug_output,
    was_connected,
    errno,
    error_msg,
    raw_data,
):
    return Response(
        MockCurlSimple(
            info={pycurl.RESPONSE_CODE: response_code},
            output=output,
            debug_output=debug_output,
            request=Request(
                # We do not need to check if token is the right one in tests:
                # 1) Library commands tests do not care about tokens. That
                #    should be covered once in a specialized test, not in every
                #    single library command test.
                # 2) If we need to test the case when a token is not accepted
                #    by pcsd, we will do so by setting an appropriate response.
                #    The actual token value doesn't matter.
                RequestTarget(label, dest_list=dest_list, token=None),
                RequestData(action, param_list, raw_data),
            ),
        ),
        was_connected=was_connected,
        errno=errno,
        error_msg=error_msg,
    )


def create_communication(
    communication_list,
    action="",
    param_list=None,
    response_code=200,
    output="",
    debug_output="",
    was_connected=True,
    errno=0,
    error_msg=None,
    raw_data=None,
):
    """
    list of dict communication_list -- each dict describes one request-response
        it accepts keys:
            string label -- required, the label of a node to talk with
            dest_list -- list of pcs.common.host.Destination where to send
                a request, defaults to [(label, default_pcsd_port)]
            string action -- pcsd url, see RequestData
            list of pairs param_list -- see RequestData
            int response_code -- http response code
            string output -- http response output
            string debug_output -- pycurl debug output
            bool was_connected -- see Response
            int errno -- see Response
            string error_msg -- see Response
            string raw_data -- see data attrib in RequestData
        if some key is not present, it is put here from common values - rest
        args of this function(except name, communication_list,
        error_msg_template)
    string action -- pcsd url, see RequestData
    list of pairs (tuple) param_list -- see RequestData
    string response_code -- http response code
    string output -- http response output
    string debug_output -- pycurl debug output
    bool was_connected -- see Response
    int errno -- see Response
    string error_msg -- see Response
    string raw_data -- see data attrib in RequestData
    """
    # We don't care about tokens, see _communication_to_response.
    common = dict(
        action=action,
        param_list=param_list if param_list else (),
        response_code=response_code,
        output=output,
        debug_output=debug_output,
        was_connected=was_connected,
        errno=errno,
        error_msg=error_msg,
        raw_data=raw_data,
    )

    response_list = []
    for communication in communication_list:
        if "dest_list" not in communication:
            communication["dest_list"] = [
                Destination(communication["label"], settings.pcsd_default_port)
            ]
        full = common.copy()
        full.update(communication)
        response_list.append(_communication_to_response(**full))

    request_list = [response.request for response in response_list]
    return request_list, response_list


def place_multinode_call(
    calls,
    name,
    node_labels=None,
    communication_list=None,
    before=None,
    **kwargs,
):
    """
    Shortcut for adding a call sending the same request to one or more nodes

    CallListBuilder calls -- list of expected calls
    string name -- the key of this call
    list node_labels -- create success responses from these nodes
    list communication_list -- use these custom responses
    **kwargs -- see __module__.create_communication
    """
    if (node_labels is None and communication_list is None) or (
        node_labels and communication_list
    ):
        raise AssertionError(
            "Exactly one of 'node_labels', 'communication_list' "
            "must be specified"
        )
    communication_list = (
        communication_list
        if communication_list is not None
        else [{"label": label} for label in node_labels]
    )
    place_communication(
        calls, name, communication_list, before=before, **kwargs
    )


def place_requests(calls, name, request_list, before=None):
    calls.place(name, AddRequestCall(request_list), before=before)


def place_responses(calls, name, response_list, before=None):
    calls.place(name, StartLoopCall(response_list), before=before)


def place_communication(calls, name, communication_list, before=None, **kwargs):
    if not communication_list:
        # If code runs a communication command with no targets specified, the
        # whole communicator and CURL machinery gets started. It doesn't
        # actually send any HTTP requests but it adds an empty list of requests
        # to CURL and starts the CURL loop. And the mock must do the same.
        place_requests(calls, f"{name}_requests", [], before=before)
        place_responses(calls, f"{name}_responses", [], before=before)
        return

    if isinstance(communication_list[0], dict):
        communication_list = [communication_list]

    request_list = []
    response_list = []
    for com_list in communication_list:
        req_list, res_list = create_communication(com_list, **kwargs)
        request_list.append(req_list)
        response_list.extend(res_list)

    place_requests(calls, f"{name}_requests", request_list[0], before=before)
    place_responses(calls, f"{name}_responses", response_list, before=before)
    for i, req_list in enumerate(request_list[1:], start=1):
        place_requests(calls, f"{name}_requests_{i}", req_list, before=before)


class AddRequestCall:
    type = CALL_TYPE_HTTP_ADD_REQUESTS

    def __init__(self, request_list):
        self.request_list = request_list

    def format(self):
        return "Requests:\n    * {0}".format(
            "\n    * ".join(
                [log_request(request) for request in self.request_list]
            )
        )

    def __repr__(self):
        return str("<HttpAddRequest '{0}'>").format(self.request_list)


class StartLoopCall:
    type = CALL_TYPE_HTTP_START_LOOP

    def format(self):
        return "Responses:\n    * {0}".format(
            "\n    * ".join(
                [log_response(response) for response in self.response_list]
            )
        )

    def __init__(self, response_list):
        self.response_list = response_list

    def __repr__(self):
        return str("<HttpStartLoop '{0}'>").format(self.response_list)


def _compare_request_data(expected, real):
    if expected == real:
        return True

    # If data is in json format, it is not possible to compare it as string,
    # because python 3 does not keep key order of dict. So if a response is
    # built by json.dumps(some_dict), the result string can vary.

    # Let's try known use: [('data_json', 'some_json_here')]
    # It means only one pair "data_json" + json string: everything else is False

    if len(expected) != 1:
        return False

    if len(real) != 1:
        return False

    if expected[0][0] != real[0][0] or expected[0][0] != "data_json":
        return False

    try:
        expected_data = json.loads(expected[0][1])
        real_data = json.loads(real[0][1])
        return expected_data == real_data
    except ValueError:
        return False


class NodeCommunicator:
    def __init__(self, call_queue=None):
        self.__call_queue = call_queue

    def add_requests(self, request_list):
        _, add_request_call = self.__call_queue.take(
            CALL_TYPE_HTTP_ADD_REQUESTS,
            request_list,
        )

        expected_request_list = add_request_call.request_list

        if len(expected_request_list) != len(request_list):
            raise different_request_lists(expected_request_list, request_list)

        errors = {}
        for i, real_request in enumerate(request_list):
            # We don't care about tokens, see _communication_to_response.
            expected_request = add_request_call.request_list[i]

            diff = {}
            if expected_request.action != real_request.action:
                diff["action"] = (expected_request.action, real_request.action)

            if expected_request.target.label != real_request.target.label:
                diff["target.label"] = (
                    expected_request.target.label,
                    real_request.target.label,
                )

            if (
                expected_request.target.dest_list
                != real_request.target.dest_list
            ):
                diff["target.dest_list"] = (
                    expected_request.target.dest_list,
                    real_request.target.dest_list,
                )

            if not _compare_request_data(
                expected_request._data.structured_data,
                real_request._data.structured_data,
            ):
                diff["data"] = (
                    expected_request._data.structured_data,
                    real_request._data.structured_data,
                )

            if diff:
                errors[i] = diff

        if errors:
            raise self.__call_queue.error_with_context(
                bad_request_list_content(errors)
            )

    def start_loop(self):
        _, call = self.__call_queue.take(CALL_TYPE_HTTP_START_LOOP)
        return call.response_list



from pcs.common.fencing_topology import (
    TARGET_TYPE_ATTRIBUTE,
    TARGET_TYPE_NODE,
    TARGET_TYPE_REGEXP,
)

__target_type_map = {
    "attrib": TARGET_TYPE_ATTRIBUTE,
    "node": TARGET_TYPE_NODE,
    "regexp": TARGET_TYPE_REGEXP,
}

target_type_map_cli_to_lib = __target_type_map

target_type_map_lib_to_cli = {
    value: key for key, value in __target_type_map.items()
}



from typing import (
    Any,
    Iterable,
    List,
    Mapping,
)

from pcs.common.const import PcmkRoleType
from pcs.common.interface.dto import to_dict
from pcs.common.pacemaker.nvset import (
    CibNvpairDto,
    CibNvsetDto,
)
from pcs.common.pacemaker.resource.operations import (
    OCF_CHECK_LEVEL_INSTANCE_ATTRIBUTE_NAME,
    CibResourceOperationDto,
)
from pcs.lib.cib.resource.types import (
    ResourceOperationIn,
    ResourceOperationOut,
)
from pcs.lib.resource_agent import (
    ResourceAgentAction,
    ResourceAgentMetadata,
)

# Operation monitor is always required, even if creating default actions was
# not requested or a resource/stonith agent does not specify it. See
# https://clusterlabs.org/pacemaker/doc/2.1/Pacemaker_Explained/html/resources.html#monitoring-resources-for-failure
_NECESSARY_OPERATIONS = {"monitor"}

_DEFAULT_INTERVALS = {"monitor": "60s"}


def get_default_operation_interval(operation_name: str) -> str:
    """
    Return default interval for given operation_name

    operation_name -- name of the operation to get the default interval for
    """
    return _DEFAULT_INTERVALS.get(operation_name, "0s")


def complete_operations_options(
    raw_operation_list: Iterable[ResourceOperationIn],
) -> List[ResourceOperationOut]:
    """
    Return a new list of operations with all necessary keys defined

    raw_operation_list -- entered operations
    """
    operation_list = []
    for raw_operation in raw_operation_list:
        operation = dict(raw_operation)
        if operation.get("interval") is None:
            operation["interval"] = get_default_operation_interval(
                str(operation["name"])
            )
        operation_list.append(operation)
    return operation_list


def get_default_operations(
    agent_metadata: ResourceAgentMetadata,
    necessary_only: bool = False,
) -> List[CibResourceOperationDto]:
    is_default_operation = (
        _is_default_operation_stonith
        if agent_metadata.name.is_stonith
        else _is_default_operation_resource
    )

    # filter actions
    action_list = [
        action
        for action in agent_metadata.actions
        if (necessary_only and action.name in _NECESSARY_OPERATIONS)
        or (not necessary_only and is_default_operation(action))
    ]

    # add necessary actions if they are missing
    defined_operation_names = frozenset(op.name for op in action_list)
    for op_name in _NECESSARY_OPERATIONS:
        if op_name not in defined_operation_names:
            action_list.append(
                ResourceAgentAction(
                    op_name, None, None, None, None, None, False, False
                )
            )

    # transform actions to operation definitions
    return [action_to_operation_dto(action) for action in action_list]


def action_to_operation_dto(
    action: ResourceAgentAction,
) -> CibResourceOperationDto:
    """
    Transform agent action data to CIB operation data
    """
    instance_attributes = []
    if action.depth not in (None, "0"):
        instance_attributes = [
            CibNvsetDto(
                id="",
                options={},
                rule=None,
                nvpairs=[
                    CibNvpairDto(
                        id="",
                        name=OCF_CHECK_LEVEL_INSTANCE_ATTRIBUTE_NAME,
                        value=str(action.depth),
                    )
                ],
            )
        ]
    return CibResourceOperationDto(
        id="",
        name=action.name,
        interval=(
            action.interval
            if action.interval
            else get_default_operation_interval(action.name)
        ),
        description=None,
        start_delay=action.start_delay,
        interval_origin=None,
        timeout=action.timeout,
        enabled=None,
        record_pending=None,
        role=PcmkRoleType(action.role) if action.role else None,
        on_fail=None,
        meta_attributes=[],
        instance_attributes=instance_attributes,
    )


def operation_dto_to_legacy_dict(
    operation: CibResourceOperationDto,
    defaults: Mapping[str, Any],
) -> ResourceOperationOut:
    operation_dict = dict(defaults)
    operation_dict.update(
        {
            key: value
            for key, value in to_dict(operation).items()
            if key in ("name", "timeout", "interval", "role")
        }
    )
    operation_dict["start-delay"] = operation.start_delay
    operation_dict[OCF_CHECK_LEVEL_INSTANCE_ATTRIBUTE_NAME] = None
    for nvset in operation.instance_attributes:
        for nvpair in nvset.nvpairs:
            if nvpair.name == OCF_CHECK_LEVEL_INSTANCE_ATTRIBUTE_NAME:
                operation_dict[
                    OCF_CHECK_LEVEL_INSTANCE_ATTRIBUTE_NAME
                ] = nvpair.value
    return operation_dict


def _is_default_operation_resource(action: ResourceAgentAction) -> bool:
    # Copy all actions to the CIB even those not defined in the OCF standard
    # or pacemaker. This way even custom actions defined in a resource agent
    # will be copied to the CIB and run by pacemaker if they specify
    # an interval. See https://github.com/ClusterLabs/pcs/issues/132
    return action.name not in {
        # one-time action, not meant to be processed by pacemaker
        "meta-data",
        # deprecated alias of monitor
        "status",
        # one-time action, not meant to be processed by pacemaker
        "validate-all",
    }


def _is_default_operation_stonith(action: ResourceAgentAction) -> bool:
    return action.name == "monitor"



import pkgutil
from typing import (
    Any,
    List,
)
from unittest import TestCase

from dataclasses import (
    dataclass,
    field,
    is_dataclass,
)

import pcs
from pcs.common.interface.dto import (
    DataTransferObject,
    from_dict,
    meta,
    to_dict,
)


def _import_all(_path):
    # arbitrary prefix so it doesn't iteract with real import in real tests
    for loader, module_name, is_pkg in pkgutil.walk_packages(
        _path, prefix="_pcs."
    ):
        if module_name.startswith("_pcs.snmp."):
            continue
        del is_pkg
        loader.find_module(module_name).load_module(module_name)


def _all_subclasses(cls):
    subclasses = set(cls.__subclasses__())
    return subclasses.union({s for c in subclasses for s in _all_subclasses(c)})


class DatatransferObjectTest(TestCase):
    def test_has_all_subclasses_are_dataclasses(self):
        _import_all(pcs.__path__)
        for cls in _all_subclasses(DataTransferObject):
            self.assertTrue(is_dataclass(cls), f"{cls} is not a dataclass")


@dataclass
class MyDto1(DataTransferObject):
    field_a: int
    field_b: int = field(metadata=meta(name="field-b"))
    field_c: int


@dataclass
class MyDto2(DataTransferObject):
    field_d: int
    field_e: MyDto1 = field(metadata=meta(name="field-e"))
    field_f: int


@dataclass
class MyDto3(DataTransferObject):
    field_g: MyDto2 = field(metadata=meta(name="field-g"))
    field_h: List[MyDto2]
    field_i: int = field(metadata=meta(name="field-i"))


class DictName(TestCase):
    simple_dto = MyDto1(1, 2, 3)
    simple_dict = {"field_a": 1, "field-b": 2, "field_c": 3}
    nested_dto = MyDto3(
        MyDto2(0, MyDto1(1, 2, 3), 4),
        [MyDto2(5, MyDto1(6, 7, 8), 9), MyDto2(10, MyDto1(11, 12, 13), 14)],
        15,
    )
    nested_dict = {
        "field-g": {
            "field_d": 0,
            "field-e": {"field_a": 1, "field-b": 2, "field_c": 3},
            "field_f": 4,
        },
        "field_h": [
            {
                "field_d": 5,
                "field-e": {"field_a": 6, "field-b": 7, "field_c": 8},
                "field_f": 9,
            },
            {
                "field_d": 10,
                "field-e": {
                    "field_a": 11,
                    "field-b": 12,
                    "field_c": 13,
                },
                "field_f": 14,
            },
        ],
        "field-i": 15,
    }

    def test_simple_to_dict(self):
        self.assertEqual(to_dict(self.simple_dto), self.simple_dict)

    def test_simple_from_dict(self):
        self.assertEqual(self.simple_dto, from_dict(MyDto1, self.simple_dict))

    def test_nested_to_dict(self):
        self.assertEqual(to_dict(self.nested_dto), self.nested_dict)

    def test_nested_from_dict(self):
        self.assertEqual(self.nested_dto, from_dict(MyDto3, self.nested_dict))


@dataclass
class DtoWithAny(DataTransferObject):
    field_a: str
    field_b: Any


class UnexpectedTypes(TestCase):
    def test_any_is_list(self):
        dto = DtoWithAny("a", [1, 2])
        self.assertEqual(dict(field_a="a", field_b=[1, 2]), to_dict(dto))

    def test_any_is_dict(self):
        dto = DtoWithAny("a", {1: "1", 2: "2"})
        self.assertEqual(
            dict(field_a="a", field_b={1: "1", 2: "2"}), to_dict(dto)
        )



import json

from pcs.common.reports import ReportItemSeverity as severities
from pcs.common.reports import codes as report_codes

ALL_RESOURCE_XML_TAGS = ["bundle", "clone", "group", "master", "primitive"]


def report_variation(report, **_info):
    updated_info = report[2].copy()
    updated_info.update(_info)
    return report[0], report[1], updated_info, report[3]


def debug(code, context=None, **kwargs):
    return severities.DEBUG, code, kwargs, None, context


def warn(code, context=None, **kwargs):
    return severities.WARNING, code, kwargs, None, context


def error(code, force_code=None, context=None, **kwargs):
    return severities.ERROR, code, kwargs, force_code, context


def info(code, context=None, **kwargs):
    return severities.INFO, code, kwargs, None, context


class ReportStore:
    def __init__(self, names=None, reports=None):
        if not names:
            names = []

        duplicate_names = {n for n in names if names.count(n) > 1}
        if duplicate_names:
            raise AssertionError(
                "Duplicate names are not allowed in ReportStore. "
                " Found duplications:\n  '{0}'".format(
                    "'\n  '".join(duplicate_names)
                )
            )

        self.__names = names
        self.__reports = reports or []
        if len(self.__names) != len(self.__reports):
            raise AssertionError("Same count reports as names required")

    @property
    def reports(self):
        return list(self.__reports)

    def adapt(self, name, **_info):
        index = self.__names.index(name)
        return ReportStore(
            self.__names,
            [
                report if i != index else report_variation(report, **_info)
                for i, report in enumerate(self.__reports)
            ],
        )

    def adapt_multi(self, name_list, **_info):
        names, reports = zip(
            *[
                (
                    name,
                    report_variation(self[name], **_info)
                    if name in name_list
                    else self[name],
                )
                for name in self.__names
            ]
        )
        return ReportStore(list(names), list(reports))

    def info(self, name, code, **kwargs):
        return self.__append(name, info(code, **kwargs))

    def warn(self, name, code, **kwargs):
        return self.__append(name, warn(code, **kwargs))

    def error(self, name, code, force_code=None, **kwargs):
        return self.__append(name, error(code, force_code=force_code, **kwargs))

    def as_warn(self, name, as_name):
        report = self[name]
        return self.__append(as_name, warn(report[1], **report[2]))

    def copy(self, name, as_name, **_info):
        return self.__append(as_name, report_variation(self[name], **_info))

    def remove(self, *name_list):
        names, reports = zip(
            *[
                (name, self[name])
                for name in self.__names
                if name not in name_list
            ]
        )
        return ReportStore(list(names), list(reports))

    def select(self, *name_list):
        names, reports = zip(*[(name, self[name]) for name in name_list])
        return ReportStore(list(names), list(reports))

    def only(self, name, **_info):
        return ReportStore([name], [report_variation(self[name], **_info)])

    def __getitem__(self, spec):
        if not isinstance(spec, slice):
            return self.__reports[self.__names.index(spec)]

        assert spec.step is None, "Step is not supported in slicing"
        start = None if spec.start is None else self.__names.index(spec.start)
        stop = None if spec.stop is None else self.__names.index(spec.stop)

        return ReportStore(self.__names[start:stop], self.__reports[start:stop])

    def __add__(self, other):
        return ReportStore(
            # pylint: disable=protected-access
            self.__names + other.__names,
            self.__reports + other.__reports,
        )

    def __append(self, name, report):
        return ReportStore(self.__names + [name], self.__reports + [report])


def report_not_found(
    res_id, context_type="", expected_types=None, context_id=""
):
    return (
        severities.ERROR,
        report_codes.ID_NOT_FOUND,
        {
            "context_type": context_type,
            "context_id": context_id,
            "id": res_id,
            "expected_types": (
                ALL_RESOURCE_XML_TAGS
                if expected_types is None
                else expected_types
            ),
        },
        None,
    )


def report_not_resource_or_tag(element_id, context_type="cib", context_id=""):
    return report_not_found(
        element_id,
        context_type=context_type,
        expected_types=sorted(ALL_RESOURCE_XML_TAGS + ["tag"]),
        context_id=context_id,
    )


def report_invalid_id(_id, invalid_char, id_description="id"):
    return (
        severities.ERROR,
        report_codes.INVALID_ID_BAD_CHAR,
        {
            "id": _id,
            "id_description": id_description,
            "is_first_char": _id.index(invalid_char) == 0,
            "invalid_character": invalid_char,
        },
        None,
    )


def report_id_already_exist(_id):
    return (
        severities.ERROR,
        report_codes.ID_ALREADY_EXISTS,
        {
            "id": _id,
        },
    )


def report_resource_not_running(resource, severity=severities.INFO):
    return (
        severity,
        report_codes.RESOURCE_DOES_NOT_RUN,
        {
            "resource_id": resource,
        },
        None,
    )


def report_resource_running(resource, roles, severity=severities.INFO):
    return (
        severity,
        report_codes.RESOURCE_RUNNING_ON_NODES,
        {
            "resource_id": resource,
            "roles_with_nodes": roles,
        },
        None,
    )


def report_unexpected_element(element_id, element_type, expected_types):
    return (
        severities.ERROR,
        report_codes.ID_BELONGS_TO_UNEXPECTED_TYPE,
        {
            "id": element_id,
            "expected_types": expected_types,
            "current_type": element_type,
        },
        None,
    )


def report_not_for_bundles(element_id):
    return report_unexpected_element(
        element_id, "bundle", ["clone", "master", "group", "primitive"]
    )


def report_wait_for_idle_timed_out(reason):
    return (
        severities.ERROR,
        report_codes.WAIT_FOR_IDLE_TIMED_OUT,
        {
            "reason": reason.strip(),
        },
        None,
    )


def check_sbd_comm_success_fixture(node, watchdog, device_list):
    return dict(
        label=node,
        output=json.dumps(
            {
                "sbd": {
                    "installed": True,
                },
                "watchdog": {
                    "exist": True,
                    "path": watchdog,
                    "is_supported": True,
                },
                "device_list": [
                    dict(path=dev, exist=True, block_device=True)
                    for dev in device_list
                ],
            }
        ),
        param_list=[
            ("watchdog", watchdog),
            ("device_list", json.dumps(device_list)),
        ],
    )



from tornado.locks import Lock

from pcs.daemon import ruby_pcsd
from pcs.daemon.app.sinatra_common import Sinatra
from pcs.daemon.auth import authorize_user
from pcs.daemon.http_server import HttpsServerManage


class SinatraRemote(Sinatra):
    """
    SinatraRemote is handler for urls which should be directed to the Sinatra
    remote (non-GUI) functions.
    """

    async def handle_sinatra_request(self):
        result = await self.ruby_pcsd_wrapper.request_remote(self.request)
        self.send_sinatra_result(result)

    async def get(self, *args, **kwargs):
        del args, kwargs
        await self.handle_sinatra_request()

    async def post(self, *args, **kwargs):
        del args, kwargs
        await self.handle_sinatra_request()


class SyncConfigMutualExclusive(SinatraRemote):
    """
    SyncConfigMutualExclusive handles urls which should be directed to the
    Sinatra remote (non-GUI) functions that can not run at the same time as
    config synchronization. The exclusivity is achieved by sync_config_lock.
    """

    def initialize(
        self, sync_config_lock: Lock, ruby_pcsd_wrapper: ruby_pcsd.Wrapper
    ):
        # pylint: disable=arguments-differ, attribute-defined-outside-init
        super().initialize(ruby_pcsd_wrapper)
        self.__sync_config_lock = sync_config_lock

    async def get(self, *args, **kwargs):
        async with self.__sync_config_lock:
            await super().get(*args, **kwargs)

    async def post(self, *args, **kwargs):
        async with self.__sync_config_lock:
            await super().get(*args, **kwargs)


class SetCerts(SinatraRemote):
    """
    SetCerts handles url for setting new certificate and key. It calls the
    Sinatra for setting certificate and key and in the case of the success it
    will take care of notify of http sever about this change.
    """

    def initialize(
        self,
        ruby_pcsd_wrapper: ruby_pcsd.Wrapper,
        https_server_manage: HttpsServerManage,
    ):
        # pylint: disable=arguments-differ, attribute-defined-outside-init
        super().initialize(ruby_pcsd_wrapper)
        self.__https_server_manage = https_server_manage

    async def handle_sinatra_request(self):
        result = await self.ruby_pcsd_wrapper.request_remote(self.request)
        if result.status == 200:
            self.__https_server_manage.reload_certs()
        self.send_sinatra_result(result)


class Auth(SinatraRemote):
    async def auth(self):
        user_auth_info = await authorize_user(
            self.get_body_argument("username"),
            self.get_body_argument("password"),
        )
        if user_auth_info.is_authorized:
            await self.handle_sinatra_request()

    async def post(self, *args, **kwargs):
        await self.auth()

    async def get(self, *args, **kwargs):
        await self.auth()


def get_routes(
    ruby_pcsd_wrapper: ruby_pcsd.Wrapper,
    sync_config_lock: Lock,
    https_server_manage: HttpsServerManage,
):
    ruby_wrapper = dict(ruby_pcsd_wrapper=ruby_pcsd_wrapper)
    lock = dict(sync_config_lock=sync_config_lock)
    server_manage = dict(https_server_manage=https_server_manage)

    return [
        # Urls protected by tokens. It is still done by ruby pcsd.
        (r"/run_pcs", SinatraRemote, ruby_wrapper),
        (r"/remote/set_certs", SetCerts, {**ruby_wrapper, **server_manage}),
        (
            r"/remote/(set_sync_options|set_configs)",
            SyncConfigMutualExclusive,
            {**ruby_wrapper, **lock},
        ),
        (r"/remote/auth", Auth, ruby_wrapper),
        (r"/remote/.*", SinatraRemote, ruby_wrapper),
        (r"/api/.*", SinatraRemote, ruby_wrapper),
    ]



from functools import partial
from unittest import TestCase

from pcs_test.tools.assertions import (
    AssertPcsMixin,
    ac,
)
from pcs_test.tools.misc import get_test_resource as rc
from pcs_test.tools.misc import (
    get_tmp_dir,
    get_tmp_file,
    skip_unless_root,
    write_file_to_tmpfile,
)
from pcs_test.tools.pcs_runner import (
    PcsRunner,
    pcs,
)


class UidGidTest(TestCase):
    # pylint: disable=invalid-name
    def setUp(self):
        self.uid_gid_dir = get_tmp_dir("tier1_cluster_uidgid")

    def tearDown(self):
        self.uid_gid_dir.cleanup()

    def testUIDGID(self):
        # pylint: disable=too-many-statements
        _pcs = partial(
            pcs,
            None,
            mock_settings={"corosync_uidgid_dir": self.uid_gid_dir.name},
        )
        o, r = _pcs("cluster uidgid".split())
        ac(o, "No uidgids configured\n")
        assert r == 0

        o, r = _pcs("cluster uidgid add".split())
        assert r == 1
        assert o.startswith("\nUsage:")

        o, r = _pcs("cluster uidgid rm".split())
        assert r == 1
        assert o.startswith(
            "Hint: This command has been replaced with 'pcs cluster uidgid "
            "delete', 'pcs cluster uidgid remove'."
        )

        o, r = _pcs("cluster uidgid xx".split())
        assert r == 1
        assert o.startswith("\nUsage:")

        o, r = _pcs("cluster uidgid add uid=testuid gid=testgid".split())
        assert r == 0
        ac(o, "")

        o, r = _pcs("cluster uidgid add uid=testuid gid=testgid".split())
        ac(
            o,
            "Error: uidgid file with uid=testuid and gid=testgid already "
            "exists\n",
        )
        assert r == 1

        o, r = _pcs("cluster uidgid delete uid=testuid2 gid=testgid2".split())
        assert r == 1
        ac(
            o,
            "Error: no uidgid files with uid=testuid2 and gid=testgid2 found\n",
        )

        o, r = _pcs("cluster uidgid remove uid=testuid gid=testgid2".split())
        assert r == 1
        ac(
            o,
            "Error: no uidgid files with uid=testuid and gid=testgid2 found\n",
        )

        o, r = _pcs("cluster uidgid rm uid=testuid2 gid=testgid".split())
        assert r == 1
        ac(
            o,
            "'pcs cluster uidgid rm' has been deprecated, use 'pcs cluster "
            "uidgid delete' or 'pcs cluster uidgid remove' instead\n"
            "Error: no uidgid files with uid=testuid2 and gid=testgid found\n",
        )

        o, r = _pcs("cluster uidgid".split())
        assert r == 0
        ac(o, "UID/GID: uid=testuid gid=testgid\n")

        o, r = _pcs("cluster uidgid delete uid=testuid gid=testgid".split())
        ac(o, "")
        assert r == 0

        o, r = _pcs("cluster uidgid add uid=testuid gid=testgid".split())
        assert r == 0
        ac(o, "")

        o, r = _pcs("cluster uidgid".split())
        assert r == 0
        ac(o, "UID/GID: uid=testuid gid=testgid\n")

        o, r = _pcs("cluster uidgid remove uid=testuid gid=testgid".split())
        ac(o, "")
        assert r == 0

        o, r = _pcs("cluster uidgid add uid=testuid gid=testgid".split())
        assert r == 0
        ac(o, "")

        o, r = _pcs("cluster uidgid".split())
        assert r == 0
        ac(o, "UID/GID: uid=testuid gid=testgid\n")

        o, r = _pcs("cluster uidgid rm uid=testuid gid=testgid".split())
        ac(
            o,
            "'pcs cluster uidgid rm' has been deprecated, use 'pcs cluster "
            "uidgid delete' or 'pcs cluster uidgid remove' instead\n",
        )
        assert r == 0

        o, r = _pcs("cluster uidgid".split())
        assert r == 0
        ac(o, "No uidgids configured\n")


class ClusterUpgradeTest(TestCase, AssertPcsMixin):
    def setUp(self):
        self.temp_cib = get_tmp_file("tier1_cluster_upgrade")
        write_file_to_tmpfile(rc("cib-empty-1.2.xml"), self.temp_cib)
        self.pcs_runner = PcsRunner(self.temp_cib.name)

    def tearDown(self):
        self.temp_cib.close()

    def test_cluster_upgrade(self):
        # pylint: disable=invalid-name
        self.temp_cib.seek(0)
        data = self.temp_cib.read()
        assert data.find("pacemaker-1.2") != -1
        assert data.find("pacemaker-2.") == -1

        o, r = pcs(self.temp_cib.name, "cluster cib-upgrade".split())
        ac(o, "Cluster CIB has been upgraded to latest version\n")
        assert r == 0

        self.temp_cib.seek(0)
        data = self.temp_cib.read()
        assert data.find("pacemaker-1.2") == -1
        assert data.find("pacemaker-2.") == -1
        assert data.find("pacemaker-3.") != -1

        o, r = pcs(self.temp_cib.name, "cluster cib-upgrade".split())
        ac(o, "Cluster CIB has been upgraded to latest version\n")
        assert r == 0


@skip_unless_root()
class ClusterStartStop(TestCase, AssertPcsMixin):
    def setUp(self):
        self.pcs_runner = PcsRunner(cib_file=None)

    def test_all_and_nodelist(self):
        self.assert_pcs_fail(
            "cluster stop rh7-1 rh7-2 --all".split(),
            "Error: Cannot specify both --all and a list of nodes.\n",
        )
        self.assert_pcs_fail(
            "cluster start rh7-1 rh7-2 --all".split(),
            "Error: Cannot specify both --all and a list of nodes.\n",
        )


@skip_unless_root()
class ClusterEnableDisable(TestCase, AssertPcsMixin):
    def setUp(self):
        self.pcs_runner = PcsRunner(cib_file=None)

    def test_all_and_nodelist(self):
        self.assert_pcs_fail(
            "cluster enable rh7-1 rh7-2 --all".split(),
            "Error: Cannot specify both --all and a list of nodes.\n",
        )
        self.assert_pcs_fail(
            "cluster disable rh7-1 rh7-2 --all".split(),
            "Error: Cannot specify both --all and a list of nodes.\n",
        )



from pcs import settings
from pcs.common import (
    file_type_codes,
    reports,
)
from pcs.common.file import RawFileError
from pcs.common.reports.item import ReportItem
from pcs.lib.errors import LibraryError

CALL_TYPE_GET_LOCAL_COROSYNC_CONF = "CALL_TYPE_GET_LOCAL_COROSYNC_CONF"


class Call:
    type = CALL_TYPE_GET_LOCAL_COROSYNC_CONF

    def __init__(self, content, exception_msg=None):
        self.content = content
        self.exception_msg = exception_msg

    def __repr__(self):
        return str("<GetLocalCorosyncConf>")


def get_get_local_corosync_conf(call_queue):
    def get_local_corosync_conf():
        _, expected_call = call_queue.take(CALL_TYPE_GET_LOCAL_COROSYNC_CONF)
        if expected_call.exception_msg:
            raise LibraryError(
                ReportItem.error(
                    reports.messages.FileIoError(
                        file_type_codes.COROSYNC_CONF,
                        RawFileError.ACTION_READ,
                        expected_call.exception_msg,
                        settings.corosync_conf_file,
                    )
                )
            )
        return expected_call.content

    return get_local_corosync_conf



from unittest import (
    TestCase,
    mock,
)

from pcs.cli.common.errors import CmdLineInputError
from pcs.cli.common.parse_args import InputModifiers
from pcs.cli.constraint_ticket import command


def _modifiers(force=True):
    options = {}
    if force:
        options["--force"] = ""
    return InputModifiers(options)


class AddTest(TestCase):
    # pylint: disable=no-self-use
    @mock.patch("pcs.cli.constraint_ticket.command.parse_args.parse_add")
    def test_call_library_with_correct_attrs(self, mock_parse_add):
        mock_parse_add.return_value = (
            "ticket",
            "resource_id",
            "",
            {"loss-policy": "fence"},
        )
        lib = mock.MagicMock()
        lib.constraint_ticket = mock.MagicMock()
        lib.constraint_ticket.create = mock.MagicMock()

        command.add(lib, ["argv"], _modifiers())

        mock_parse_add.assert_called_once_with(["argv"])
        lib.constraint_ticket.create.assert_called_once_with(
            "ticket",
            "resource_id",
            {"loss-policy": "fence"},
            resource_in_clone_alowed=True,
            duplication_alowed=True,
        )

    @mock.patch("pcs.cli.constraint_ticket.command.parse_args.parse_add")
    def test_refuse_resource_role_in_options(self, mock_parse_add):
        mock_parse_add.return_value = (
            "ticket",
            "resource_id",
            "resource_role",
            {"rsc-role": "master"},
        )
        lib = None
        self.assertRaises(
            CmdLineInputError, lambda: command.add(lib, ["argv"], _modifiers())
        )

    @mock.patch("pcs.cli.constraint_ticket.command.parse_args.parse_add")
    def test_put_resource_role_to_options_for_library(self, mock_parse_add):
        mock_parse_add.return_value = (
            "ticket",
            "resource_id",
            "resource_role",
            {"loss-policy": "fence"},
        )
        lib = mock.MagicMock()
        lib.constraint_ticket = mock.MagicMock()
        lib.constraint_ticket.create = mock.MagicMock()

        command.add(lib, ["argv"], _modifiers())

        mock_parse_add.assert_called_once_with(["argv"])
        lib.constraint_ticket.create.assert_called_once_with(
            "ticket",
            "resource_id",
            {"loss-policy": "fence", "rsc-role": "resource_role"},
            resource_in_clone_alowed=True,
            duplication_alowed=True,
        )


class RemoveTest(TestCase):
    def test_refuse_args_count(self):
        self.assertRaises(
            CmdLineInputError,
            lambda: command.remove(
                mock.MagicMock(),
                ["TICKET"],
                _modifiers(False),
            ),
        )
        self.assertRaises(
            CmdLineInputError,
            lambda: command.remove(
                mock.MagicMock(),
                ["TICKET", "RESOURCE", "SOMETHING_ELSE"],
                _modifiers(False),
            ),
        )

    def test_call_library_remove_with_correct_attrs(self):
        # pylint: disable=no-self-use
        lib = mock.MagicMock(
            constraint_ticket=mock.MagicMock(remove=mock.Mock())
        )
        command.remove(lib, ["TICKET", "RESOURCE"], _modifiers(False))
        lib.constraint_ticket.remove.assert_called_once_with(
            "TICKET",
            "RESOURCE",
        )



from typing import (
    List,
    Optional,
)

from dataclasses import dataclass

from pcs.common.interface.dto import DataTransferObject


@dataclass(frozen=True)
class ServiceStatusDto(DataTransferObject):
    service: str
    installed: Optional[bool]
    enabled: Optional[bool]
    running: Optional[bool]


@dataclass(frozen=True)
class ServicesInfoResultDto(DataTransferObject):
    services: List[ServiceStatusDto]



def fixture_resources_for_ids():
    return """
        <resources>
            <primitive class="ocf" id="id1" provider="pacemaker" type="Dummy"/>
            <primitive class="ocf" id="id2" provider="pacemaker" type="Dummy"/>
        </resources>
    """


def fixture_tags_xml(tag_with_ids):
    return (
        "<tags>"
        + "".join(
            '<tag id="{tag_id}">'.format(tag_id=tag_id)
            + "".join('<obj_ref id="{id}"/>'.format(id=id) for id in id_list)
            + "</tag>"
            for tag_id, id_list in tag_with_ids
        )
        + "</tags>"
    )


def fixture_constraints_for_tags(tag_list):
    return (
        "<constraints>"
        + "".join(
            f"<rsc_location id='location-{tag}' rsc='{tag}' node='nodeN' "
            "score='100'/>"
            for tag in tag_list
        )
        + "</constraints>"
    )


def fixture_resources_for_reference_ids(id_list):
    return (
        "<resources>"
        + "".join(
            (
                f"<primitive class='ocf' id='{_id}' provider='pacemaker' "
                "type='Dummy'/>"
            )
            for _id in id_list
        )
        + "</resources>"
    )



from typing import (
    NewType,
    Optional,
    Sequence,
)

from dataclasses import dataclass

from pcs.common.interface.dto import DataTransferObject
from pcs.common.pacemaker.nvset import CibNvsetDto

ContainerType = NewType("ContainerType", str)

CONTAINER_TYPE_DOCKER = ContainerType("docker")
CONTAINER_TYPE_PODMAN = ContainerType("podman")
CONTAINER_TYPE_RKT = ContainerType("rkt")


@dataclass(frozen=True)
class CibResourceBundleContainerRuntimeOptionsDto(DataTransferObject):
    image: str
    replicas: Optional[int]
    replicas_per_host: Optional[int]
    promoted_max: Optional[int]
    run_command: Optional[str]
    network: Optional[str]
    options: Optional[str]


@dataclass(frozen=True)
class CibResourceBundlePortMappingDto(DataTransferObject):
    id: str  # pylint: disable=invalid-name
    port: Optional[int]
    internal_port: Optional[int]
    range: Optional[str]


@dataclass(frozen=True)
class CibResourceBundleNetworkOptionsDto(DataTransferObject):
    ip_range_start: Optional[str]
    control_port: Optional[int]
    host_interface: Optional[str]
    host_netmask: Optional[int]
    add_host: Optional[bool]


@dataclass(frozen=True)
class CibResourceBundleStorageMappingDto(DataTransferObject):
    id: str  # pylint: disable=invalid-name
    source_dir: Optional[str]
    source_dir_root: Optional[str]
    target_dir: str
    options: Optional[str]


@dataclass(frozen=True)
class CibResourceBundleDto(DataTransferObject):
    # pylint: disable=too-many-instance-attributes
    id: str  # pylint: disable=invalid-name
    description: Optional[str]
    member_id: Optional[str]
    container_type: Optional[ContainerType]
    container_options: Optional[CibResourceBundleContainerRuntimeOptionsDto]
    network: Optional[CibResourceBundleNetworkOptionsDto]
    port_mappings: Sequence[CibResourceBundlePortMappingDto]
    storage_mappings: Sequence[CibResourceBundleStorageMappingDto]
    meta_attributes: Sequence[CibNvsetDto]
    instance_attributes: Sequence[CibNvsetDto]



from typing import (
    Iterable,
    List,
    Optional,
)

from lxml.etree import _Element

from pcs.common import reports
from pcs.common.reports.const import (
    ADD_REMOVE_CONTAINER_TYPE_GROUP,
    ADD_REMOVE_ITEM_TYPE_RESOURCE,
)
from pcs.common.reports.item import (
    ReportItem,
    ReportItemList,
)
from pcs.lib.cib.resource import (
    clone,
    group,
)
from pcs.lib.cib.resource.common import (
    get_inner_resources,
    get_parent_resource,
    is_resource,
    is_wrapper_resource,
)
from pcs.lib.validate import validate_add_remove_items


def validate_move_resources_to_group(
    group_element: _Element,
    resource_element_list: List[_Element],
    adjacent_resource_element: Optional[_Element],
) -> ReportItemList:
    """
    Validates that existing resources can be moved into a group,
    optionally beside an adjacent_resource_element

    group_element -- the group to put resources into
    resource_element_list -- resources that are being moved into the group
    adjacent_resource_element -- put resources beside this one
    """
    report_list: ReportItemList = []

    # Validate types of resources and their parents
    for resource_element in resource_element_list:
        # Only primitive resources can be moved
        if not is_resource(resource_element):
            report_list.append(
                ReportItem.error(
                    reports.messages.IdBelongsToUnexpectedType(
                        str(resource_element.attrib["id"]),
                        ["primitive"],
                        resource_element.tag,
                    )
                )
            )
        elif is_wrapper_resource(resource_element):
            report_list.append(
                ReportItem.error(
                    reports.messages.CannotGroupResourceWrongType(
                        str(resource_element.attrib["id"]),
                        resource_element.tag,
                        parent_id=None,
                        parent_type=None,
                    )
                )
            )
        else:
            parent = get_parent_resource(resource_element)
            if parent is not None and not group.is_group(parent):
                # At the moment, moving resources out of bundles and clones
                # (or masters) is not possible
                report_list.append(
                    ReportItem.error(
                        reports.messages.CannotGroupResourceWrongType(
                            str(resource_element.attrib["id"]),
                            resource_element.tag,
                            parent_id=str(parent.attrib["id"]),
                            parent_type=parent.tag,
                        )
                    )
                )

    # Validate that elements can be added
    # Check if the group element is a group
    if group.is_group(group_element):
        report_list += validate_add_remove_items(
            [str(resource.attrib["id"]) for resource in resource_element_list],
            [],
            [
                str(resource.attrib["id"])
                for resource in get_inner_resources(group_element)
            ],
            ADD_REMOVE_CONTAINER_TYPE_GROUP,
            ADD_REMOVE_ITEM_TYPE_RESOURCE,
            str(group_element.attrib["id"]),
            str(adjacent_resource_element.attrib["id"])
            if adjacent_resource_element is not None
            else None,
            True,
        )
    else:
        report_list.append(
            ReportItem.error(
                reports.messages.IdBelongsToUnexpectedType(
                    str(group_element.attrib["id"]),
                    expected_types=[group.TAG],
                    current_type=group_element.tag,
                )
            )
        )

    # Elements can always be removed from their old groups, except when the last
    # resource is removed but that is handled in resource.group_add for now, no
    # need to run the validation for removing elements
    return report_list


def move_resources_to_group(
    group_element: _Element,
    primitives_to_place: Iterable[_Element],
    adjacent_resource: Optional[_Element] = None,
    put_after_adjacent: bool = True,
) -> None:
    """
    Put resources into a group or move them within their group

    There is a corner case which is not covered in this function. If the CIB
    contains references to a group or clone which this function deletes,
    they are not deleted and an invalid CIB is generated. These references
    can be constraints, fencing levels etc. - anything that contains group id of
    the deleted group. It is on the caller to detect this corner case and handle
    it appropriately (see group_add in lib/commands/resource.py). For future
    rewrites of this function, it would be better to ask for --force before
    deleting anything that user didn't explicitly ask for - like deleting the
    clone and its associated constraints.

    etree.Element group_element -- the group to put resources into
    iterable primitives_to_place -- resource elements to put into the group
    etree.Element adjacent_resource -- put resources beside this one if set
    bool put_after_adjacent -- put resources after or before the adjacent one
    """
    for resource in primitives_to_place:
        old_parent = resource.getparent()

        # Move a resource to the group.
        if (
            adjacent_resource is not None
            and adjacent_resource.getnext() is not None
            and put_after_adjacent
        ):
            adjacent_resource.getnext().addprevious(resource)  # type: ignore
            adjacent_resource = resource
        elif adjacent_resource is not None and not put_after_adjacent:
            adjacent_resource.addprevious(resource)
        else:
            group_element.append(resource)
            adjacent_resource = resource

        # If the resource was the last resource in another group, that group is
        # now empty and must be deleted. If the group is in a clone element,
        # delete that as well.
        if (
            old_parent is not None
            and group.is_group(old_parent)  # do not delete resources element
            and not group.get_inner_resources(old_parent)
        ):
            old_grandparent = old_parent.getparent()
            if old_grandparent is not None:
                old_great_grandparent = old_grandparent.getparent()
                if (
                    clone.is_any_clone(old_grandparent)
                    and old_great_grandparent is not None
                ):
                    old_great_grandparent.remove(old_grandparent)
                else:
                    old_grandparent.remove(old_parent)



from unittest import TestCase

from lxml import etree

from pcs_test.tools.cib import get_assert_pcs_effect_mixin
from pcs_test.tools.misc import get_test_resource as rc
from pcs_test.tools.misc import (
    get_tmp_file,
    write_file_to_tmpfile,
)
from pcs_test.tools.pcs_runner import PcsRunner


def get_cib_resources(cib):
    return etree.tostring(etree.parse(cib).findall(".//resources")[0])


class ResourceTest(TestCase, get_assert_pcs_effect_mixin(get_cib_resources)):
    empty_cib = rc("cib-empty.xml")

    def setUp(self):
        self.temp_cib = get_tmp_file("tier1_test_resource_common")
        write_file_to_tmpfile(self.empty_cib, self.temp_cib)
        self.pcs_runner = PcsRunner(self.temp_cib.name)

    def tearDown(self):
        self.temp_cib.close()



from pcs import usage
from pcs.cli.common.routing import create_router
from pcs.cli.tag import command as tag

tag_cmd = create_router(
    {
        "config": tag.tag_config,
        "create": tag.tag_create,
        "delete": tag.tag_remove,
        "help": lambda lib, argv, modifiers: usage.tag(argv),
        # TODO remove, deprecated command
        # replaced with 'config'
        "list": tag.tag_list_cmd,
        "remove": tag.tag_remove,
        "update": tag.tag_update,
    },
    ["tag"],
    default_cmd="config",
)
